"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import imp
import json
import logging
import os
import os.path
from time import sleep

import redis
from redis.exceptions import BusyLoadingError, RedisError

from agora.engine.utils import prepare_store_path

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.utils.kv')


def __check_kv(kv):
    # type: () -> None
    reqs = 0
    while True:
        log.debug('Checking Redis... ({})'.format(reqs))
        reqs += 1
        try:
            kv.echo('echo')
            break
        except BusyLoadingError as e:
            log.warning(e.message)
        except RedisError, e:
            log.error('Redis is not available')
            raise e
    return kv


kvs = []
path_kvs = {}


def get_redis_lite(*args, **kwargs):
    try:
        imp.find_module('redislite')
        import redislite

        if args:
            settings_file = args[0] + '.settings'

            try:
                remove_settings = False
                with open(settings_file, 'r') as f:
                    settings = json.load(f)
                    if not os.path.exists(settings['unixsocket']):
                        remove_settings = True
            except Exception as e:
                pass

            if remove_settings:
                os.remove(settings_file)

        return redislite.StrictRedis(*args, **kwargs)


    except ImportError:
        raise EnvironmentError('Redislite module is not available')


def get_kv(persist_mode=None, redis_host='localhost', redis_port=6379, redis_db=1, redis_file=None, base='store',
           path='', **kwargs):
    if persist_mode is None or persist_mode:
        if redis_file is not None:
            prepare_store_path(base, path)
            redis_path = str('/'.join(filter(lambda x: x, [base, path, redis_file])))
            kv = get_redis_lite(redis_path)
            kvs.append(kv)
            path_kvs[redis_path + '.settings'] = kv
            return kv
        else:
            pool = redis.ConnectionPool(host=redis_host, port=redis_port, db=redis_db)
            kv = redis.StrictRedis(connection_pool=pool)
            return __check_kv(kv)
    else:
        return get_redis_lite()


def close():
    for r in kvs:
        kvs.remove(r)
        close_kv_connection(r)


def close_kv_connection(r):
    retries = 0
    while retries < 3:
        try:
            r.save()
            r.shutdown()
            return True
        except Exception as e:
            if 'connecting to unix socket' in e.message:
                break
            retries += 1
            if retries == 3:
                log.error(e.message)
                break
            sleep(0.1)


def close_kv(kv, clear=False):
    if isinstance(kv, str):
        if kv in path_kvs:
            path_kv = kv
            kv = path_kvs[path_kv]
            del path_kvs[path_kv]
        else:
            return False

    if kv in kvs:
        kvs.remove(kv)
        close_kv_connection(kv)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import logging

import networkx as nx
from networkx import Graph

from agora.engine.fountain.index import Index
from agora.engine.fountain.seed import SeedManager

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.fountain.path')

match_elm_cycles = {}
paths_cache = {}


def _build_directed_graph(index, generic=False, graph=None):
    # type: (Index, bool, nx.DiGraph) -> Graph

    def dom_edge():
        d_cons = p_dict['constraints'].get(d, None)
        data = {'c': []}
        if d_cons is not None:
            data['c'] = d_cons
        return d, node, data

    if graph is None:
        graph = nx.DiGraph()
    else:
        graph.clear()

    paths_cache.clear()

    graph.add_nodes_from(index.types, ty='type')
    for node in index.properties:
        p_dict = index.get_property(node)
        if p_dict['type'] != 'object':
            continue

        dom = set(p_dict.get('domain'))
        dom = filter(lambda x: index.get_type(x)['spec_refs'] or node in index.get_type(x)['spec_properties'], dom)
        ran = set(p_dict.get('range'))
        try:
            ran = filter(lambda x: node in index.get_type(x)['spec_refs'] or index.get_type(x)['spec_properties'],
                         ran)
        except TypeError:
            pass
        edges = []
        edges.extend([dom_edge() for d in dom])
        if p_dict.get('type') == 'object':
            edges.extend([(node, r) for r in ran])
        graph.add_edges_from(edges)
        graph.add_node(node, ty='prop', object=p_dict.get('type') == 'object', range=ran,
                       constraints=p_dict['constraints'])

    log.debug('Link graph edges: {}'.format(len(graph.edges())))
    return graph


def _find_cycles(index):
    # type: (Index) -> iter

    g_graph = _build_directed_graph(index, generic=True)
    cycle_keys = index.r.keys('*cycles*')
    for ck in cycle_keys:
        index.r.delete(ck)
    g_cycles = list(nx.simple_cycles(g_graph))
    with index.r.pipeline() as pipe:
        pipe.multi()
        for i, cy in enumerate(g_cycles):
            cycle = []
            t_cycle = None
            cycle_types = set()
            for j, elm in enumerate(cy):
                if index.is_type(elm):
                    t_cycle = elm
                elif t_cycle is not None:
                    cons = g_graph.get_edge_data(t_cycle, elm)['c']
                    if cons:
                        next_type = cy[(j + 1) % len(cy)]
                        if next_type not in cons:
                            t_cycle = None
                            cycle = []
                            cycle_types.clear()
                            break

                    cycle.append({'property': elm, 'type': t_cycle})
                    cycle_types.add(t_cycle)
                    t_cycle = None
            if t_cycle is not None:
                cons = g_graph.get_edge_data(t_cycle, cy[0])['c']
                if cons:
                    next_type = cy[1]
                    if next_type not in cons:
                        t_cycle = None
                        cycle = []
                        cycle_types.clear()

                if t_cycle:
                    cycle.append({'property': cy[0], 'type': t_cycle})
                    cycle_types.add(t_cycle)
            if cycle:
                pipe.zadd('cycles', i, cycle)
            cycle_types = reduce(lambda x, y: x.union(set(index.get_type(y)['sub'] + [y])), cycle_types, set())
            for ct in cycle_types:
                pipe.sadd('cycles:{}'.format(ct), i)
        pipe.execute()
    return g_cycles


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in xrange(0, len(l), n):
        yield l[i:i + n]


def _all_simple_paths(graph, source, target):
    if (graph, source, target) not in paths_cache:
        paths_cache[(graph, source, target)] = list(nx.all_simple_paths(graph, source, target))
    return paths_cache.get((graph, source, target), [])


def _get_simple_paths(index, graph, source, target):
    if (graph, source, target) not in paths_cache:
        paths = _all_simple_paths(graph, source, target)
        source_type_dict = index.get_type(source)
        target_type_dict = index.get_type(target)
        source_type_super = source_type_dict['super']
        target_type_super = target_type_dict['super']

        if not paths and not source_type_super:
            source_type_super.append(source)

        for ss_ty in source_type_super:
            paths.extend(_all_simple_paths(graph, ss_ty, target))
            for ts_ty in target_type_super:
                paths.extend(_all_simple_paths(graph, source, ts_ty))
                sp = _all_simple_paths(graph, ss_ty, ts_ty)
                paths.extend(sp)
        final_paths = []
        for p in paths:
            if p not in final_paths:
                final_paths.append(p)
        if not final_paths:
            raise nx.NetworkXNoPath
        paths_cache[(graph, source, target)] = final_paths
    return paths_cache[(graph, source, target)]


def _find_seed_paths(index, sm, graph, elm, force_seed=None):
    # elm must always be a type
    if index.is_property(elm):
        targets = index.get_property(elm)['domain']
        type = False
    elif index.is_type(elm):
        targets = [elm]
        targets.extend(index.get_type(elm)['sub'])
        type = True
    else:
        raise TypeError('{}?'.format(elm))

    targets = filter(lambda t: not set.intersection(set(index.get_type(t)['super']), targets) or
                               index.get_type(t)['spec_refs'], targets)

    if force_seed:
        seed_types = map(lambda x: x[1], force_seed)
    else:
        seed_types = sm.seeds.keys()

    seed_types = filter(lambda t: not set.intersection(set(index.get_type(t)['super']), seed_types) or
                                  index.get_type(t)['spec_properties'], seed_types)

    yielded = []
    for ty in seed_types:
        for target in targets:
            try:
                ty_paths = _get_simple_paths(index, graph, ty, target)
            except nx.NetworkXNoPath:
                if ty == target:
                    yield ty, []
                    continue
                else:
                    ty_paths = None

            if ty_paths is not None:
                if ty == target or target in index.get_type(ty)['super']:
                    path = []
                    if not type:
                        path.append({'type': ty, 'property': elm})
                    if (ty, path) not in yielded:
                        yielded.append((ty, path))
                        yield ty, path
                for path in ty_paths:
                    path = map(lambda x: {'type': x[0], 'property': x[1]}, chunks(path[:-1], 2))
                    last_step = None
                    filtering = False
                    n_iterations = len(path)
                    if type:
                        n_iterations += 1
                    for i in range(n_iterations):
                        step_type = path[i]['type'] if i < len(path) else elm
                        if last_step is not None:
                            source = graph.get_edge_data(last_step['type'], last_step['property'])
                            link_constraints = source.get('c', [])
                            if link_constraints:
                                step_types = [step_type] + index.get_type(step_type)['sub']
                                if not set.intersection(set(step_types), set(link_constraints)):
                                    filtering = True
                                    break
                        last_step = path[i] if i < len(path) else last_step
                    if not filtering:
                        if not type:
                            path.append({'type': target, 'property': elm})
                        if (ty, path) not in yielded:
                            yielded.append((ty, path))
                            yield ty, path


def _find_path(index, sm, graph, elm, force_seed=None):
    # type: (Index, SeedManager, nx.DiGraph, str, tuple) -> tuple

    def find_seeds_for(ty):
        if force_seed:
            for s in [x[0] for x in force_seed if x[1] == ty or x[1] in index.get_type(ty)['sub']]:
                yield s
        else:
            for s in sm.get_type_seeds(ty):
                yield s

    def find_property_paths(elm):
        return list(_find_seed_paths(index, sm, graph, elm, force_seed=force_seed))

    def build_seed_path(ty, path):
        type_seeds = list(find_seeds_for(ty))
        if len(type_seeds):
            return {'cycles': [], 'seeds': type_seeds, 'steps': path}

    paths = find_property_paths(elm)
    seed_paths = filter(lambda x: x is not None, map(lambda (ty, path): build_seed_path(ty, path), paths))

    cycle_ids = set([int(c) for c in index.r.smembers('cycles:{}'.format(elm))])

    for path in seed_paths:
        path['cycles'] = cycle_ids.copy()
        for step in path['steps']:
            cycles = set([int(c) for c in index.r.smembers('cycles:{}'.format(step.get('type')))])
            path['cycles'].update(set(path['cycles']).union(cycles))
            cycle_ids.update(cycles)
        path['cycles'] = list(path['cycles'])

    applying_cycles = set(cycle_ids)

    new_applying_cycles = {}
    for cid in applying_cycles:
        try:
            eval_cycle = eval(index.r.zrange('cycles', cid, cid).pop())
            new_applying_cycles[int(cid)] = eval_cycle
        except:
            pass

    applying_cycles = new_applying_cycles

    # applying_cycles = {int(cid): eval(index.r.zrange('cycles', cid, cid).pop()) for cid in applying_cycles}
    return seed_paths, [{'cycle': cid, 'steps': applying_cycles[cid]} for cid in
                        applying_cycles]


class PathManager(object):
    def __init__(self):
        self.__index = None
        self.__sm = None
        self.__pgraph = nx.DiGraph()
        self.__last_ts = -1

    @property
    def index(self):
        return self.__index

    @index.setter
    def index(self, i):
        self.__index = i
        _build_directed_graph(self.index, graph=self.__pgraph)

    @property
    def seed_manager(self):
        return self.__sm

    @seed_manager.setter
    def seed_manager(self, s):
        self.__sm = s

    def calculate(self):
        _build_directed_graph(self.__index, graph=self.__pgraph)
        _find_cycles(self.__index)

    def __check_graph(self):
        current_ts = self.__index.ts
        if self.__last_ts < current_ts:
            self.calculate()
            self.__last_ts = current_ts

    def get_paths(self, elm, force_seed=None):
        self.__check_graph()
        seed_paths, all_cycles = _find_path(self.__index, self.__sm, self.__pgraph, elm, force_seed=force_seed)
        return {'paths': seed_paths, 'all-cycles': all_cycles}

    @property
    def path_graph(self):
        self.__check_graph()
        return self.__pgraph

    def are_connected(self, source, target):
        self.__check_graph()
        try:
            return bool(_get_simple_paths(self.index, self.path_graph, source, target))
        except nx.NetworkXNoPath:
            return False



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import itertools
from agora.collector.remote import RemoteCollector
from agora.graph import AgoraGraph
from agora.server.fountain import client as fc
from agora.server.planner import client as pc

from agora.server.fragment import client as cc
from rdflib import Graph

__author__ = 'Fernando Serena'


class AgoraClient(object):
    def __init__(self, fountain_host='localhost', fountain_port=5000, planner_host='localhost', planner_port=5000,
                 fragment_host='localhost', fragment_port=5000):
        self._fountain = fc(host=fountain_host, port=fountain_port)
        self._planner = pc(host=planner_host, port=planner_port, fountain=self._fountain)
        self._collector = RemoteCollector(fragment_host, fragment_port, planner=self._planner)
        self._fragment = cc(host=fragment_host, port=fragment_port)

    @property
    def fountain(self):
        return self._fountain

    @property
    def planner(self):
        return self._planner

    def query(self, query, **kwargs):
        graph = AgoraGraph(self._collector)
        return graph.query(query, **kwargs)

    def fragment(self, query=None, agps=None):
        if not (query or agps):
            return

        graph = AgoraGraph(self._collector)
        result = Graph(namespace_manager=graph.namespace_manager)

        if query:
            for c, s, p, o in self._fragment.fragment(query):
                result.add((s, p, o))
        else:
            for agp in agps:
                for c, s, p, o in graph.collector.get_fragment_generator(agp)['generator']:
                    result.add((s, p, o))
        return result

    def fragment_generator(self, query=None, agps=None):
        def comp_gen(gens):
            for gen in [g['generator'] for g in gens]:
                for q in gen:
                    yield q

        if query is not None:
            generator = self._fragment.fragment(query)
            plan = self.search_plan(query)
        else:
            graph = AgoraGraph(self._collector)
            generators = [graph.collector.get_fragment_generator(agp) for agp in agps]
            plan = Graph(namespace_manager=graph.namespace_manager)
            for g in generators:
                plan.__iadd__(g['plan'])
            generator = comp_gen(generators)

        prefixes = dict(plan.namespaces())

        return {'prefixes': prefixes, 'plan': plan, 'generator': generator}

    def search_plan(self, query):
        graph = AgoraGraph(self._collector)
        comp_plan = Graph(namespace_manager=graph.namespace_manager)
        for agp in graph.agps(query):
            comp_plan.__iadd__(self._planner.make_plan(agp))
        return comp_plan

    def agps(self, query):
        graph = AgoraGraph(self._collector)
        return graph.agps(query)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import re
from StringIO import StringIO

from rdflib import Graph, URIRef, BNode

from agora.engine.fountain import AbstractFountain
from agora.engine.plan import AGP
from agora.engine.plan import AbstractPlanner
from agora.server import Server, TURTLE, HTML, Client, APIError, JSON, tuples_force_seed, dict_force_seed
from agora.server.fountain import client as fc

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.server.planner')


def skolemize(g):
    def _skolem(n):
        if isinstance(n, BNode):
            if n not in bnodes_map:
                bnodes_map[n] = URIRef('http://bnodes/{}'.format(str(n)))
            n = bnodes_map[n]
        return n

    bnodes_map = {}
    skolem = Graph()
    for s, p, o in g:
        s = _skolem(s)
        o = _skolem(o)
        skolem.add((s, p, o))
    return skolem


def deskolemize(g):
    def _deskolem(n):
        if isinstance(n, URIRef) and n.startswith('http://bnodes'):
            if n not in bnodes_map:
                bnodes_map[n] = BNode(n.replace('http://bnodes/', ''))
            n = bnodes_map[n]
        return n

    bnodes_map = {}
    deskolem = Graph()
    for s, p, o in g:
        s = _deskolem(s)
        o = _deskolem(o)
        deskolem.add((s, p, o))
    return deskolem


def make_plan(planner, agp_str, force_seed):
    try:
        agp_str = agp_str.lstrip('{').rstrip('}').strip()
        tps = re.split('\. ', agp_str)
        tps = map(lambda x: x.rstrip('.').strip(), filter(lambda y: y != '', tps))
        agp = AGP(tps, prefixes=planner.fountain.prefixes)
        plan = planner.make_plan(agp, force_seed=force_seed)
        plan = skolemize(plan)
        return plan.serialize(format='turtle')
    except NameError, e:
        raise APIError(e.message)


def build(planner, server=None, import_name=__name__):
    # type: (AbstractPlanner, Server, str) -> Server

    if server is None:
        server = Server(import_name)

    @server.get('/plan', produce_types=(TURTLE, HTML))
    def make_plan_get():
        # type: () -> str
        agp_str = server.request_args.get('agp', '{}')
        return make_plan(planner, agp_str, None)

    @server.post('/plan', produce_types=(TURTLE, HTML), consume_types=(JSON,))
    def make_plan_force_seeds(req_json):
        # type: (dict) -> str
        force_seed = list(tuples_force_seed(req_json['force_seed']))
        return make_plan(planner, req_json['agp'], force_seed)

    return server


class PlannerClient(Client, AbstractPlanner):
    def __init__(self, host='localhost', port=5000, fountain=None):
        # type: (str, int, AbstractFountain) -> PlannerClient
        super(PlannerClient, self).__init__(host, port)
        self.__fountain = fc(host, port) if fountain is None else fountain

    def make_plan(self, agp, force_seed=None):
        # type: (AGP, dict) -> Graph
        if force_seed:
            url = 'plan'
            type_seed_dict = dict_force_seed(force_seed)
            req_json = {'agp': str(agp), 'force_seed': type_seed_dict}
            response = self._post_request(url, req_json, content_type='application/json', accept='text/turtle')
        else:
            url = 'plan?agp=%s' % agp
            response = self._get_request(url, accept='text/turtle')

        graph = Graph()
        graph.parse(StringIO(response), format='text/turtle')
        graph = deskolemize(graph)

        return graph

    @property
    def fountain(self):
        # type: () -> AbstractFountain
        return self.__fountain


def client(host='localhost', port=5000, fountain=None):
    # type: (str, int, AbstractFountain) -> PlannerClient
    return PlannerClient(host, port, fountain)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import os
from threading import Lock

__author__ = 'Fernando Serena'


class Singleton(type):
    _lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            id = kwargs.get('id', '')
            if id not in cls.instances:
                cls.instances[id] = super(Singleton, cls).__call__(*args, **kwargs)
            return cls.instances[id]


def prepare_store_path(base, path):
    if base and not os.path.exists(base):
        os.makedirs(base)
    if base or path:
        full_path = '/'.join(filter(lambda x: x, [base, path]))
        if not os.path.exists(full_path):
            os.makedirs(full_path)
        return full_path


class Semaphore(object):
    def __init__(self):
        self.value = -1

    def __enter__(self):
        self.value = 0

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.value = 1

    def isSet(self):
        return self.value == 1

    def set(self):
        self.value = 1


def get_immediate_subdirectories(a_dir):
    if not os.path.exists(a_dir):
        os.makedirs(a_dir)
    return [name for name in os.listdir(a_dir)
            if os.path.isdir(os.path.join(a_dir, name))]


class Wrapper(object):
    def __init__(self, object):
        self.__lock = Lock()
        self.__attr_locks = {}

        if isinstance(object, Wrapper):
            self.__object = object.__object
            self.__argret = object.__argret
            self.__attr_locks = object.__attr_locks
            self.__lock = object.__lock
        else:
            self.__object = object
            self.__argret = {}

    def decorate(self, f):
        def wrapper(*args, **kwargs):

            key = f.func_name + str(args)
            with self.__lock:
                if key not in self.__attr_locks:
                    self.__attr_locks[key] = Lock()

            with self.__attr_locks[key]:
                if key not in self.__argret:
                    ret = f(*args, **kwargs)

                    if kwargs:
                        return ret
                    if ret is not None:
                        self.__argret[key] = ret

                return self.__argret.get(key, None)

        return wrapper

    def __getattr__(self, item):

        if item.startswith('__'):
            return self.__object.__getattribute__(item)

        with self.__lock:
            if item not in self.__attr_locks:
                self.__attr_locks[item] = Lock()

        with self.__attr_locks[item]:
            if item not in self.__argret:
                attr = self.__object.__getattribute__(item)
                if hasattr(attr, '__call__'):
                    return self.decorate(attr)
                else:
                    self.__argret[item] = attr

        return self.__argret[item]


stopped = Semaphore()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging

from rdflib.plugins.sparql.algebra import translateQuery
from rdflib.plugins.sparql.parser import parseQuery
from rdflib.plugins.sparql.processor import SPARQLResult, SPARQLProcessor
from rdflib.plugins.sparql.sparql import Query
from rdflib.query import ResultRow

from agora.graph.evaluate import evalQuery, traverse_part

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.graph.processor')


class FragmentResult(SPARQLResult):
    def __init__(self, res):
        super(FragmentResult, self).__init__(res)
        self.chunk_size = res.get('chunk_', None)

    def __iter__(self):
        if self.type in ("CONSTRUCT", "DESCRIBE"):
            for t in self.graph:
                yield t
        elif self.type == 'ASK':
            yield self.askAnswer
        elif self.type == 'SELECT':
            # this iterates over ResultRows of variable bindings
            if self._genbindings:
                for b in self._genbindings:
                    self._bindings.append(b)
                    yield ResultRow(b, self.vars)
                self._genbindings = None
            else:
                for b in self._bindings:
                    yield ResultRow(b, self.vars)


class FragmentProcessor(SPARQLProcessor):
    def query(
            self, strOrQuery, initBindings={},
            initNs={}, base=None, DEBUG=False, **kwargs):
        """
        Evaluate a query with the given initial bindings, and initial
        namespaces. The given base is used to resolve relative URIs in
        the query and will be overridden by any BASE given in the query.
        """

        if not isinstance(strOrQuery, Query):
            parsetree = parseQuery(strOrQuery)
            query = translateQuery(parsetree, base, initNs)
        else:
            query = strOrQuery

        collector = kwargs.get('collector', None)
        graph = kwargs.get('graph', None)
        if collector is not None and graph is not None:
            part = query.algebra
            filters = {}
            bgps = []

            for p in traverse_part(part, filters):
                bgps.append(p)

            incremental = False
            for bgp in bgps:
                filters = {v: filters[v] for v in bgp._vars if v in filters}
                f = collector.get(graph.build_agp(bgp.triples), filters)
                if f is None or not f['fragment'].updated:
                    incremental = True
                    break
            kwargs['incremental'] = incremental

        del kwargs['collector']
        del kwargs['graph']
        eval = evalQuery(self.graph, query, initBindings, base, **kwargs)
        return eval



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import StringIO
import logging
import urlparse

from rdflib import Graph, RDF, URIRef
from rdflib.namespace import OWL
from rdflib.plugins.parsers.notation3 import BadSyntax

from agora.engine.fountain.schema import Schema

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.fountain.onto')


class VocabularyError(Exception):
    pass


class DuplicateVocabulary(VocabularyError):
    pass


class VocabularyNotFound(VocabularyError):
    pass


class UnknownVocabulary(VocabularyError):
    pass


def __load_owl(owl):
    """

    :param owl: The ontology to be loaded to the fountain
    :return:
    """
    raw_g = Graph()
    for f in ['turtle', 'xml']:
        try:
            raw_g.parse(source=StringIO.StringIO(owl), format=f)
            log.debug('Parsed ontology in {} format'.format(f))
            break
        except SyntaxError, e:
            log.warn(e.message)
            pass

    if not len(raw_g):
        raise VocabularyError()

    owl_g = Graph()
    owl_g.bind('owl', OWL)
    for s, p, o in raw_g.triples((None, None, None)):
        if (isinstance(s, URIRef) and s.startswith('file')) or p.startswith('file') or (
                    isinstance(o, URIRef) and o.startswith('file')):
            continue
        owl_g.add((s, p, o))

    found_ontos = list(owl_g.subjects(RDF.type, OWL.Ontology))
    if len(found_ontos) != 1:
        raise VocabularyNotFound("Incorrect number of ontology statements: {}".format(len(found_ontos)))

    uri = found_ontos.pop()

    for p, u in raw_g.namespaces():
        if not p.startswith('ns'):
            owl_g.bind(p, u)

    vid = [p for (p, u) in owl_g.namespaces() if uri in u and p != '']
    imports = owl_g.objects(uri, OWL.imports)
    if not len(vid):
        vid = urlparse.urlparse(uri).path.split('/')[-1]
    else:
        vid = vid.pop()

    # (identifier, ontology uri, graph, imports)
    return vid, uri, owl_g, imports


def add_vocabulary(schema, owl):
    # type: (Schema, str) -> iter
    vid, uri, owl_g, imports = __load_owl(owl)

    if vid in schema.contexts:
        raise DuplicateVocabulary('Vocabulary already contained')

    schema.add_context(vid, owl_g)
    vids = [vid]

    # Add imported vocabularies
    for im_uri in imports:
        log.debug('Importing {} from {}...'.format(im_uri, vid))
        im_g = Graph()
        try:
            im_g.load(im_uri, format='turtle')
        except BadSyntax:
            try:
                im_g.load(im_uri)
            except BadSyntax:
                log.error('Bad syntax in {}'.format(im_uri))
        except Exception:
            log.warn("Couldn't import {}".format(im_uri))
        else:
            try:
                child_vids = add_vocabulary(schema, im_g.serialize(format='turtle'))
                vids.extend(child_vids)
            except DuplicateVocabulary, e:
                log.debug('vocabulary already added: {}'.format(im_uri))
            except VocabularyNotFound, e:
                log.warning('uri not found for {}'.format(im_uri))
            except Exception, e:
                log.error(e.message)

    return vids


def update_vocabulary(schema, vid, owl):
    """

    :param vid:
    :param owl:
    :return:
    """
    owl_vid, uri, owl_g, imports = __load_owl(owl)

    if vid != owl_vid:
        raise Exception("Identifiers don't match")

    if vid not in schema.contexts:
        raise UnknownVocabulary('Vocabulary id is not known')

    schema.update_context(vid, owl_g)


def delete_vocabulary(schema, vid):
    if vid not in schema.contexts:
        raise UnknownVocabulary('Vocabulary id is not known')

    schema.remove_context(vid)


def get_vocabularies(schema):
    """

    :return:
    """
    return schema.contexts


def get_vocabulary(schema, vid):
    """

    :param vid:
    :return:
    """
    return schema.get_context(vid).serialize(format='turtle')



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
from base64 import b64encode
from functools import wraps

from rdflib import ConjunctiveGraph
from rdflib import Graph

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.utils.cache')


class Cache(dict):
    """
    Special dictionary to be used as a cache that offers mechanisms to watch and clear (on cascade)
    """

    def __init__(self, **kwargs):
        super(Cache, self).__init__(**kwargs)
        self.__watchers = []
        self.stable = 1
        self.ts = -1

    def watch(self, other):
        # type: (Cache) -> None
        if isinstance(other, Cache):
            if self not in other.__watchers:
                other.__watchers.append(self)

    def clear(self):
        # type: () -> None
        super(Cache, self).clear()
        for obs in self.__watchers:
            obs.clear()


def cached(cache, level=0, ref_ts=0):
    # type: (Cache, int, int) -> callable

    if cache.ts < ref_ts:
        cache.clear()
        cache.ts = ref_ts

    def d(f):
        @wraps(f)
        def wrap(*args, **kwargs):
            if cache is not None:
                if not isinstance(cache, Cache):
                    raise AttributeError('Cache type is not valid')
                cache_key = b64encode(f.__name__ + str(args[1:]) + str(kwargs))
                if not cache.stable >= level or cache_key not in cache:
                    result = f(*args, **kwargs)
                    cache[cache_key] = result
                return cache[cache_key]
            else:
                return f(*args, **kwargs)

        return wrap

    return d


class SubGraph(Graph):
    def __init__(self, cache, store='default', identifier=None,
                 namespace_manager=None):
        super(SubGraph, self).__init__(store, identifier, namespace_manager)
        self.__cache = cache

    def query(self, q, **kwargs):
        if self.__cache is None:
            return set(super(SubGraph, self).query(q))

        key = b64encode(q + self.identifier)
        if key not in self.__cache:
            r = set(super(SubGraph, self).query(q))
            self.__cache[key] = r
        return self.__cache[key]


class ContextGraph(ConjunctiveGraph):
    def __init__(self, cache, store='default', identifier=None):
        # type: (Cache, str, str) -> ContextGraph
        super(ContextGraph, self).__init__(store, identifier)
        self.__cache = cache

    def query(self, q, **kwargs):
        if self.__cache is None:
            return set(super(ContextGraph, self).query(q))

        key = b64encode(q + self.identifier)
        if key not in self.__cache:
            r = set(super(ContextGraph, self).query(q))
            self.__cache[key] = r
        return self.__cache[key]

    def remove_context(self, context):
        super(ContextGraph, self).remove_context(context)
        self.__cache.clear()

    def get_context(self, identifier, quoted=False):
        # type: (any, bool) -> SubGraph
        return SubGraph(self.__cache, store=self.store, identifier=identifier,
                        namespace_manager=self.namespace_manager)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2017 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import signal
from multiprocessing import Lock

from rdflib import Graph

from agora.collector import Collector
from agora.collector.cache import RedisCache
from agora.engine.fountain import Fountain
from agora.engine.fountain.index import Index
from agora.engine.fountain.path import PathManager
from agora.engine.fountain.schema import Schema
from agora.engine.fountain.seed import SeedManager
from agora.engine.plan import Planner, AbstractPlanner
from agora.engine.utils import stopped, Wrapper
from agora.engine.utils.graph import get_cached_triple_store
from agora.engine.utils.kv import get_kv, close, close_kv
from agora.graph import AgoraGraph
from agora.server.fountain import FountainClient
from agora.server.fountain import client as fc
from agora.server.planner import client as pc, PlannerClient

__author__ = 'Fernando Serena'


def signal_term_handler(signal, frame):
    print 'got SIGTERM'
    Agora.close()
    # sys.exit(0)


signal.signal(signal.SIGTERM, signal_term_handler)


def setup_logging(level):
    log = logging.getLogger('agora')
    log.setLevel(level)
    for h in log.handlers:
        log.removeHandler(h)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setLevel(level)
    ch.setFormatter(formatter)
    log.addHandler(ch)


class Agora(object):
    def __init__(self, **kwargs):
        self.__lock = Lock()

    @property
    def fountain(self):
        return self.planner.fountain

    @property
    def planner(self):
        # type: () -> AbstractPlanner
        return self._planner

    @planner.setter
    def planner(self, p):
        self._planner = p

    def __get_agora_graph(self, collector, cache, loader, force_seed):
        if collector is None:
            collector = Collector()
            collector.planner = self.planner
            collector.cache = cache
        if loader is not None:
            collector.loader = loader
        if force_seed is not None:
            collector.force_seed = force_seed
        return AgoraGraph(collector)

    def query(self, query, collector=None, cache=None, loader=None, force_seed=None, **kwargs):
        graph = self.__get_agora_graph(collector, cache, loader, force_seed)
        return graph.query(query, collector=collector, cache=cache, loader=loader, force_seed=force_seed, **kwargs)

    def fragment(self, query=None, agps=None, collector=None, cache=None, loader=None, force_seed=None):
        if not (query or agps):
            return

        graph = self.__get_agora_graph(collector, cache, loader, force_seed)
        result = Graph(namespace_manager=graph.namespace_manager)

        agps = list(graph.agps(query)) if query else agps

        for agp in agps:
            for c, s, p, o in graph.collector.get_fragment_generator(agp)['generator']:
                result.add((s, p, o))
        return result

    def fragment_generator(self, query=None, agps=None, collector=None, cache=None, loader=None, force_seed=None,
                           stop_event=None, follow_cycles=True):
        def comp_gen(gens):
            for gen in [g['generator'] for g in gens]:
                for q in gen:
                    yield q

        graph = self.__get_agora_graph(collector, cache, loader, force_seed)
        agps = list(graph.agps(query)) if query else agps

        generators = [graph.collector.get_fragment_generator(agp, filters=filters, stop_event=stop_event,
                                                             follow_cycles=follow_cycles) for
                      agp, filters in
                      agps]
        prefixes = {}
        comp_plan = Graph(namespace_manager=graph.namespace_manager)
        for g in generators:
            comp_plan.__iadd__(g['plan'])
            prefixes.update(g['prefixes'])

        return {'prefixes': prefixes, 'plan': comp_plan, 'generator': comp_gen(generators), 'gens': generators}

    def search_plan(self, query, force_seed=None):
        collector = Collector()
        collector.planner = self.planner
        graph = AgoraGraph(collector)
        comp_plan = Graph(namespace_manager=graph.namespace_manager)
        for agp, filters in graph.agps(query):
            comp_plan.__iadd__(self._planner.make_plan(agp, force_seed))
        return comp_plan

    def agp(self, query):
        collector = Collector()
        collector.planner = self.planner
        graph = AgoraGraph(collector)
        return graph.agps(query)

    def __new__(cls, **kwargs):
        a = super(Agora, cls).__new__(cls)
        auto = kwargs.get('auto', True)
        if not auto:
            return a

        planner = None
        fountain_host = kwargs.get('fountain_host', None)
        planner_host = kwargs.get('planner_host', None)

        if fountain_host is None and 'persist_mode' not in kwargs:
            kwargs['persist_mode'] = False

        if fountain_host is None and planner_host is None:
            kv_args = kwargs.copy()
            kv_args['path'] = ''
            kv = get_kv(**kv_args)

            cached_schema = kwargs.get('cached_schema', None)
            schema = Schema(cache=cached_schema if cached_schema is not None else True)
            if 'cached_schema' in kwargs:
                del kwargs['cached_schema']

            index = Index()
            index.r = kv

            try:
                schema.graph = get_cached_triple_store(schema.cache, **kwargs)
            except EnvironmentError as e:
                index.r.flushdb()
                raise e

            index.schema = schema
            sm = SeedManager()
            sm.index = index
            pm = PathManager()
            pm.index = index
            pm.seed_manager = sm

            fountain = Fountain()
            fountain.schema = schema
            fountain.index = index
            fountain.seed_manager = sm
            fountain.path_manager = pm
            planner = Planner(fountain)
        elif planner_host:
            client_args = {'host': planner_host}
            planner_port = kwargs.get('planner_port')
            if planner_port:
                client_args['port'] = planner_port
            planner = PlannerClient(**client_args)
        elif fountain_host:
            client_args = {'host': fountain_host}
            fountain_port = kwargs.get('fountain_port', None)
            if fountain_port:
                client_args['port'] = fountain_port
            fountain = FountainClient(**client_args)
            planner = Planner(fountain)

        if planner:
            a.planner = planner

        return a

    def shutdown(self):
        with self.__lock:
            fountain = self.fountain
            if hasattr(fountain, 'index'):
                index = fountain.index
                index.close()
            if hasattr(fountain, 'schema'):
                schema = fountain.schema
                schema.close()

    @staticmethod
    def close():
        close()
        stopped.set()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import logging

import rdflib
from rdflib import BNode
from rdflib import ConjunctiveGraph
from rdflib import Graph
from rdflib import Literal
from rdflib import RDF
from rdflib import RDFS
from rdflib import URIRef
from rdflib import Variable
from rdflib import plugin

from agora.engine.plan.agp import TP, AGP
from agora.engine.plan.graph import AGORA

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.graph')

plugin.register('agora', rdflib.query.Processor, 'agora.graph.processor', 'FragmentProcessor')
plugin.register('agora', rdflib.query.Result, 'agora.graph.processor', 'FragmentResult')


def extract_tps_from_plan(plan):
    # type: (Graph) -> dict
    """

    :param plan: Search Plan graph
    :return: A string triple representing the pattern for all triple pattern nodes
    """

    def extract_node_id(node):
        nid = node
        if (node, RDF.type, AGORA.Variable) in plan:
            nid = Variable(list(plan.objects(node, RDFS.label)).pop())
        elif (node, RDF.type, AGORA.Literal) in plan:
            nid = Literal(list(plan.objects(node, AGORA.value)).pop())
        return nid

    def process_tp_node(tpn):
        predicate = list(plan.objects(tpn, AGORA.predicate)).pop()
        subject_node = list(plan.objects(tpn, AGORA.subject)).pop()
        object_node = list(plan.objects(tpn, AGORA.object)).pop()
        subject = extract_node_id(subject_node)
        object = extract_node_id(object_node)

        return TP(subject, predicate, object)

    return {str(list(plan.objects(tpn, RDFS.label)).pop()): process_tp_node(tpn) for tpn in
            plan.subjects(RDF.type, AGORA.TriplePattern)}


def extract_seed_types_from_plan(plan):
    # type: (Graph) -> dict
    so = list(plan.subject_objects(predicate=AGORA.fromType))
    st_dict = {t: set(plan.objects(subject=tree, predicate=AGORA.hasSeed)) for tree, t in so}
    return st_dict


class AgoraGraph(ConjunctiveGraph):
    def __init__(self, collector):
        super(AgoraGraph, self).__init__()
        self.__collector = collector
        self.__collected = {}
        for prefix, ns in collector.prefixes.items():
            self.bind(prefix, ns)

    def build_agp(self, bgp):
        def tp_part(term):
            if isinstance(term, Variable) or isinstance(term, BNode):
                return '?{}'.format(str(term))
            elif isinstance(term, URIRef):
                return '<{}>'.format(term)
            elif isinstance(term, Literal):
                return term.n3(namespace_manager=self.namespace_manager)

        tps = set([])
        for s, p, o in bgp:
            s_elm = tp_part(s)
            if p == RDF.type and isinstance(o, URIRef):
                o_elm = self.qname(o)
                p_elm = 'a'
            else:
                p_elm = self.qname(p)
                o_elm = tp_part(o)

            tps.add('{} {} {}'.format(s_elm, p_elm, o_elm))
        return AGP(tps, self.__collector.prefixes)

    def gen(self, bgp, filters=None, **kwargs):
        bgp = frozenset(bgp)
        if bgp not in self.__collected:
            agp = self.build_agp(bgp)
            gen_dict = self.__collector.get_fragment_generator(agp, filters=filters, **kwargs)
            return self._produce(gen_dict, bgp)
        else:
            return self._read(bgp)

    def _read(self, bgp):
        for c, tp in self.__collected[bgp]['contexts'].values():
            for s, p, o in self.get_context(c):
                yield tp, s, p, o

    def _produce(self, gen, bgp):
        contexts = {}
        for context, s, p, o in gen['generator']:
            log.debug('Got triple: {} {} {} .'.format(s.encode('utf8', 'replace'), p.encode('utf8', 'replace'),
                                                      o.encode('utf8', 'replace')))
            context_id = str(context)
            if context_id not in contexts:
                contexts[context_id] = (BNode(context_id), context)
            blank_context = contexts.get(context_id)[0]
            self.get_context(blank_context).add((s, p, o))
            yield context, s, p, o

        self.__collected[bgp] = {'plan': gen['plan'], 'contexts': contexts}

    @property
    def collected(self):
        return iter(self.__collected)

    @property
    def collector(self):
        return self.__collector

    def load(self, *tps):
        for x in self.gen(*tps):
            print x

    def query(self, query_object, **kwargs):
        result = super(AgoraGraph, self).query(query_object, processor='agora',
                                               result='agora', graph=self, **kwargs)

        return result

    def search_plan(self, query_object, **kwargs):
        result = super(AgoraGraph, self).query(query_object, processor='agora',
                                               result='agora', **kwargs)

        return result.plan

    def agps(self, query_object):
        from agora.graph.evaluate import extract_bgps
        for bgp, filters in extract_bgps(query_object, prefixes=self.__collector.prefixes):
            yield self.build_agp(bgp.triples), filters



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import calendar
import hashlib
import logging
import traceback
from Queue import Empty, Full, Queue
from StringIO import StringIO
from datetime import datetime, timedelta
from multiprocessing import cpu_count
from threading import Event, Lock, Thread
from time import sleep

import networkx as nx
import redis
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait
from rdflib import ConjunctiveGraph, Graph
from rdflib import Literal, RDF, RDFS, URIRef, Variable
from rdflib.plugins.sparql.algebra import translateQuery
from rdflib.plugins.sparql.parser import Query
from rdflib.plugins.sparql.parser import expandUnicodeEscapes
from rdflib.plugins.sparql.sparql import QueryContext
from shortuuid import uuid

from agora.collector import Collector, triplify
from agora.collector.execution import StopException
from agora.collector.plan import FilterTree
from agora.engine.plan.agp import TP, AGP
from agora.engine.plan.graph import AGORA
from agora.engine.utils import stopped, Singleton, Semaphore
from agora.engine.utils.graph import get_triple_store
from agora.engine.utils.kv import get_kv, close_kv
from agora.graph import extract_tps_from_plan, extract_seed_types_from_plan

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.collector.scholar')


def _remove_tp_filters(tp, filter_mapping={}, prefixes=None):
    # type: (TP, dict, dict) -> (str, dict)
    """
    Transforms a triple pattern that may contain filters to a new one with both subject and object bounded
    to variables
    :param tp: The triple pattern to be filtered
    :return: Filtered triple pattern + filter mapping
    """

    def __create_var(elm):
        if elm in filter_mapping.values():
            elm = list(filter(lambda x: filter_mapping[x] == elm, filter_mapping)).pop()
        else:
            v = Variable('v' + uuid())
            filter_mapping[v] = elm
            elm = v
        return elm

    s, p, o = tp
    if isinstance(s, URIRef):
        s = __create_var(s)
    if not isinstance(o, Variable):
        o = __create_var(o)
    return TP(s, p, o)


def _generalize_agp(agp, prefixes=None):
    # Create a filtered graph pattern from the request one (general_gp)
    general_gp = AGP(prefixes=prefixes)
    filter_mapping = {}
    for new_tp in map(lambda x: _remove_tp_filters(x, filter_mapping, prefixes=prefixes), agp):
        general_gp.add(new_tp)
    return general_gp, filter_mapping


class FragmentStream(object):
    def __init__(self, store, key):
        # type: (redis.StrictRedis, str) -> None
        self.key = key
        self.store = store

    def get(self, until):
        # type: (int) -> iter
        if until is None:
            until = '+inf'
        for x in self.store.zrangebyscore(self.key, '-inf', '{}'.format(float(until))):
            yield triplify(x)

    def put(self, tp, (s, p, o), timestamp=None):
        try:
            if timestamp is None:
                timestamp = calendar.timegm(datetime.utcnow().timetuple())
            quad = (tp, s.n3(), p.n3(), o.n3())
            not_found = not bool(self.store.zscore(self.key, quad))
            if not_found:
                with self.store.pipeline() as pipe:
                    pipe.zadd(self.key, timestamp, quad)
                    pipe.execute()
            return not_found
        except Exception as e:
            traceback.print_exc()
            log.error(e.message)
            raise e

    def clear(self):
        with self.store.pipeline() as pipe:
            pipe.delete(self.key)
            pipe.execute()


class Fragment(object):
    def __init__(self, agp, kv, triples, fragments_key, fid, filters=None, follow_cycles=True):
        self.__lock = Lock()
        self.key = '{}:{}'.format(fragments_key, fid)
        self.__agp = agp
        self.__fragments_key = fragments_key
        self.fid = fid
        self.kv = kv
        self.triples = triples
        self.__stream = FragmentStream(kv, '{}:stream'.format(self.key))
        self.__plan = None
        self.__plan_event = Event()
        self.__plan_event.clear()
        self.__stop_event = Semaphore()
        self.__updated = False
        self.__aborted = False
        self.__tp_map = {}
        self.__seed_types = {}
        self.__observers = set([])
        self.collecting = False
        self.__filters = filters if isinstance(filters, dict) else {}
        self.__follow_cycles = follow_cycles

    @property
    def lock(self):
        return self.__lock

    @property
    def stream(self):
        return self.__stream

    @property
    def agp(self):
        return self.__agp

    @property
    def filters(self):
        return self.__filters

    @property
    def seed_types(self):
        return frozenset(self.__seed_types)

    @property
    def seed_digests(self):
        return self.__seed_digests

    @property
    def updated(self):
        #     with self.lock:
        self.__updated = False if self.kv.get('{}:updated'.format(self.key)) is None else True
        return self.__updated

    @property
    def follow_cycles(self):
        return self.__follow_cycles

    @property
    def updated_ts(self):
        # with self.lock:
        upd_ts = self.kv.get('{}:updated'.format(self.key)) or 0
        upd_dt = datetime.utcfromtimestamp(upd_ts)
        return upd_dt

    @property
    def newcomer(self):
        # with self.lock:
        return self.kv.get('{}:stored'.format(self.key)) is None

    def updated_for(self, ttl):
        ttl = int(min(10000000, ttl))
        ttl = int(max(ttl, 1))
        self.__updated = ttl > 0
        updated_key = '{}:updated'.format(self.key)
        with self.kv.pipeline() as pipe:
            if self.__updated:
                pipe.set(updated_key, ttl)
                pipe.set('{}:stored'.format(self.key), True)
                pipe.expire(updated_key, int(min(10000000, ttl)))
            else:
                pipe.delete(updated_key)
            pipe.execute()
        log.info('Fragment {} will be up-to-date for {}s'.format(self.fid, ttl))

    @property
    def collecting(self):
        # with self.lock:
        return False if self.kv.get('{}:collecting'.format(self.key)) is None else True

    @property
    def aborted(self):
        return self.__aborted

    @property
    def stored(self):
        # with self.lock:
        return False if self.kv.get('{}:stored'.format(self.key)) is None else True

    @collecting.setter
    def collecting(self, state):
        collecting_key = '{}:collecting'.format(self.key)
        if state:
            self.kv.set(collecting_key, state)
        else:
            self.kv.delete(collecting_key)

    @classmethod
    def load(cls, kv, triples, fragments_key, fid, prefixes=None):
        # type: (redis.StrictRedis, ConjunctiveGraph, str, str) -> Fragment
        try:
            if not any([fid in c.identifier for c in list(triples.contexts())]):
                raise EnvironmentError('Fragment context is not present in the triple store')

            agp = AGP(kv.smembers('{}:{}:gp'.format(fragments_key, fid)), prefixes=prefixes)
            plan_turtle = kv.get('{}:{}:plan'.format(fragments_key, fid))
            follow_cycles = bool(int(kv.get('{}:{}:fc'.format(fragments_key, fid))))
            fragment = Fragment(agp, kv, triples, fragments_key, fid, follow_cycles=follow_cycles)
            fragment.plan = Graph().parse(StringIO(plan_turtle), format='turtle')
            sparql_filter_keys = kv.keys('{}:{}:filters:*'.format(fragments_key, fid))
            for var_filter_key in sparql_filter_keys:
                v = var_filter_key.split(':')[-1]
                fragment.filters[Variable(v)] = set(kv.smembers(var_filter_key))
            return fragment
        except Exception, e:
            with kv.pipeline() as pipe:
                for fragment_key in kv.keys('{}*{}*'.format(fragments_key, fid)):
                    pipe.delete(fragment_key)
                pipe.execute()

    def save(self, pipe):
        fragment_key = '{}:{}'.format(self.__fragments_key, self.fid)
        pipe.delete(fragment_key)
        pipe.sadd('{}:gp'.format(fragment_key), *self.__agp)
        pipe.set('{}:fc'.format(fragment_key), 1 if self.__follow_cycles else 0)
        for v in self.__filters.keys():
            for f in self.__filters[v]:
                pipe.sadd('{}:filters:{}'.format(fragment_key, str(v)), f)

    @property
    def generator(self):
        def w_listen(ts):
            def listen(quad):
                if datetime.utcnow() > ts:
                    try:
                        listen_queue.put_nowait(quad)
                    except Full as e:
                        log.warn(e.message)
                    except Exception:
                        traceback.print_exc()

            return listen

        if self.stored:
            for c in self.__tp_map:
                for s, p, o in self.triples.get_context(str((self.fid, self.__tp_map[c]))):
                    yield self.__tp_map[c], s, p, o
        else:
            until = datetime.utcnow()
            listener = w_listen(until)
            try:
                until_ts = calendar.timegm(until.timetuple())
                listen_queue = Queue(maxsize=10000)
                with self.__lock:
                    self.__observers.add(listener)

                for c, s, p, o in self.stream.get(until_ts):
                    yield self.__tp_map[c], s, p, o

                while not self.__aborted and (not self.__updated or not listen_queue.empty()):
                    try:
                        c, s, p, o = listen_queue.get(timeout=0.1)
                        yield self.__tp_map[c], s, p, o
                    except Empty:
                        pass
                    except KeyboardInterrupt:
                        stopped.set()
                    except Exception:
                        traceback.print_exc()
                        break
            finally:
                with self.__lock:
                    self.__observers.remove(listener)

    @property
    def plan(self):
        # type: () -> Graph
        self.__plan_event.wait()
        return self.__plan

    @plan.setter
    def plan(self, p):
        # type: (Graph) -> None
        self.__plan = p
        with self.kv.pipeline() as pipe:
            g = Graph()
            plan_str = p.skolemize(g).serialize(format='turtle')
            pipe.set('{}:plan'.format(self.key), plan_str)
            pipe.execute()
        self.__tp_map = extract_tps_from_plan(self.__plan)
        self.__seed_types = extract_seed_types_from_plan(self.__plan)
        self.__calculate_seed_digests()
        self.__plan_event.set()

    def __calculate_seed_digests(self):
        self.__seed_digests = {}
        for type, seeds in self.__seed_types.items():
            m = hashlib.md5()
            for seed in sorted(seeds):
                m.update(seed)
            self.__seed_digests[type] = m.digest().encode('base64').strip()

    def __notify(self, quad):
        with self.__lock:
            for observer in self.__observers:
                observer(quad)

    def shutdown(self):
        self.__stop_event.set()

    def populate(self, collector):
        self.collecting = True
        self.stream.clear()

        completed = False
        n_triples = 0

        try:
            collect_dict = collector.get_fragment_generator(self.agp, filters=self.filters,
                                                            stop_event=self.__stop_event,
                                                            follow_cycles=self.follow_cycles)
            generator = collect_dict['generator']
            self.plan = collect_dict['plan']
            self.__aborted = False
        except Exception:
            self.__plan_event.set()
            self.__aborted = True
        else:
            back_id = uuid()
            pre_time = datetime.utcnow()
            try:
                while not completed:
                    c, s, p, o = generator.next()
                    tp = self.__tp_map[c.id]
                    self.stream.put(c.id, (s, p, o))
                    self.triples.get_context(str((back_id, tp))).add((s, p, o))
                    self.__notify((c.id, s, p, o))
                    n_triples += 1
            except StopIteration:
                completed = True
            except StopException:
                self.__aborted = True
            except Exception:
                traceback.print_exc()
                self.__aborted = True

        try:
            with self.lock:
                if not stopped.isSet() and completed and not self.aborted:
                    # Replace graph store and update ttl
                    for tp in self.__tp_map.values():
                        self.triples.remove_context(self.triples.get_context(str((self.fid, tp))))
                        self.triples.get_context(str((self.fid, tp))).__iadd__(
                            self.triples.get_context(str((back_id, tp))))
                        self.triples.remove_context(self.triples.get_context(str((back_id, tp))))
                    actual_ttl = collect_dict.get('ttl')()
                    if not n_triples and actual_ttl >= 10000000:
                        actual_ttl = 0
                    elapsed = (datetime.utcnow() - pre_time).total_seconds()
                    fragment_ttl = max(actual_ttl, elapsed)
                    self.updated_for(fragment_ttl)
                    log.info(
                        'Finished fragment collection: {} ({} triples), in {}s'.format(self.fid, n_triples, elapsed))
                elif self.aborted:
                    self.remove()
        except Exception, e:
            traceback.print_exc()
        finally:
            self.collecting = False

    def remove(self):
        # type: () -> None
        # Clear stream
        self.__stop_event.set()
        self.stream.clear()

        # Remove graph contexts
        if self.__tp_map:
            for tp in self.__tp_map.values():
                self.triples.remove_context(self.triples.get_context(str((self.fid, tp))))

        # Remove fragment keys in kv
        with self.kv.pipeline() as pipe:
            for fragment_key in self.kv.keys('{}*{}*'.format(self.__fragments_key, self.fid)):
                pipe.delete(fragment_key)
            pipe.execute()

    def mapping(self, agp, filters):
        # type: (AGP, dict) -> dict
        agp_map = self.agp.mapping(agp)
        if agp_map:
            if not self.filters:
                return agp_map
            rev_map = {v: k for k, v in agp_map.items()}
            filter_map = {rev_map[v]: set(map(lambda x: x.replace(v, rev_map[v]), fs)) for v, fs in filters.items() if
                          v in rev_map}
            # There may be incoming filters that do not affect the agp
            if filter_map == {v: f for v, f in self.filters.items() if v in rev_map}:
                return agp_map


class FragmentIndex(object):
    __metaclass__ = Singleton
    instances = {}
    daemon_event = Event()
    daemon_event.clear()
    daemon_th = None
    tpool = ThreadPoolExecutor(max_workers=cpu_count())

    def __init__(self, id='', **kwargs):
        # type: (any, str) -> FragmentIndex
        self.__orphaned = {}
        self.__key_prefix = id
        self.__fragments_key = '{}:fragments'.format(id)
        self.lock = Lock()
        # Load fragments from kv
        self.__fragments = dict(self.__load_fragments())
        self.__id = id

    def __new__(cls, id='', force_seed=None, **kwargs):
        index = super(FragmentIndex, cls).__new__(cls)
        planner = kwargs.get('planner', None)
        index.planner = planner
        loader = kwargs.get('loader', None)
        index.loader = loader
        cache = kwargs.get('cache', None)
        index.cache = cache
        index.force_seed = force_seed
        triples = get_triple_store(**kwargs)
        if cache is not None:
            kv = cache.r
        else:
            kv = get_kv(**kwargs)
        index.triples = triples
        index.kv = kv

        if FragmentIndex.daemon_th is None:
            FragmentIndex.daemon_th = Thread(target=FragmentIndex._daemon)
            FragmentIndex.daemon_th.start()

        return index

    def clear(self):
        with self.lock:
            for fragment in self.fragments.values():
                fragment.remove()
            self.__fragments.clear()

    def shutdown(self):
        with self.lock:
            for fragment in self.__fragments.values():
                fragment.shutdown()

            sleep(1)

            for fragment in self.__fragments.values():
                fragment.remove()

            self.__fragments.clear()

            del FragmentIndex.instances[self.id]
            try:
                close_kv(self.kv)
            except Exception:
                pass

    def notify(self):
        FragmentIndex.daemon_event.set()

    @property
    def id(self):
        return self.__id

    @property
    def kv(self):
        return self.__kv

    @kv.setter
    def kv(self, k):
        self.__kv = k

    @property
    def planner(self):
        return self.__planner

    @planner.setter
    def planner(self, p):
        self.__planner = p

    @property
    def cache(self):
        return self.__cache

    @cache.setter
    def cache(self, c):
        self.__cache = c

    @property
    def triples(self):
        return self.__triples

    @triples.setter
    def triples(self, t):
        self.__triples = t

    @property
    def loader(self):
        return self.__loader

    @loader.setter
    def loader(self, l):
        self.__loader = l

    @property
    def force_seed(self):
        return self.__force_seed

    @force_seed.setter
    def force_seed(self, s):
        self.__force_seed = s

    def seed_type_digest(self, ty):
        if not self.force_seed:
            return self.planner.fountain.get_seed_type_digest(ty)
        else:
            m = hashlib.md5()
            for seed in sorted(self.force_seed[ty]):
                m.update(seed)
            current_digest = m.digest().encode('base64').strip()
            return current_digest

    def __load_fragments(self):
        fids = self.kv.smembers(self.__fragments_key)
        orph_fids = self.kv.smembers('{}:orph'.format(self.__fragments_key))

        for fragment_id in fids:
            fragment = Fragment.load(self.kv, self.triples, self.__fragments_key, fragment_id,
                                     prefixes=self.planner.fountain.prefixes)

            if fragment_id in orph_fids:
                self.kv.srem(self.__fragments_key, fragment_id)
                self.kv.srem('{}:orph'.format(self.__fragments_key), fragment_id)
                if fragment is not None:
                    fragment.remove()
                continue

            if fragment is not None:
                with fragment.lock:
                    # All those fragments that were not fully collected are marked here to be orphaned
                    if not fragment.updated and not fragment.collecting:
                        self.kv.set('{}:stored'.format(fragment.key), False)
                    else:
                        yield (fragment_id, fragment)
            else:
                self.kv.srem(self.__fragments_key, fragment_id)
                self.kv.srem('{}:orph'.format(self.__fragments_key), fragment_id)

    def get(self, agp, general=False, filters=None):
        # type: (AGP, bool) -> dict
        mapping = None
        fragment = None
        filter_mapping = {}
        with self.lock:
            active_fragments = filter(lambda x: x not in self.__orphaned, self.__fragments.keys())
            for fragment_id in sorted(active_fragments, key=lambda x: abs(
                    len(self.__fragments[x].filters) - len(filters)), reverse=False):
                fragment = self.__fragments[fragment_id]
                mapping = fragment.mapping(agp, filters)
                if mapping:
                    break

            if general and mapping is None:
                general, filter_mapping = _generalize_agp(agp, prefixes=self.__planner.fountain.prefixes)
                for fragment_id in active_fragments:
                    mapping = self.__fragments[fragment_id].mapping(general, {})
                    if mapping:
                        break

            if mapping is not None:
                return {'fragment': fragment, 'vars': mapping, 'filters': filter_mapping}

    @property
    def fragments(self):
        return self.__fragments

    def register(self, agp, filters=None, follow_cycles=True):
        # type: (AGP, dict) -> Fragment
        with self.lock:
            fragment_id = str(uuid())
            fragment = Fragment(agp, self.kv, self.triples, self.__fragments_key, fragment_id, filters=filters,
                                follow_cycles=follow_cycles)
            with self.kv.pipeline() as pipe:
                pipe.sadd(self.__fragments_key, fragment_id)
                fragment.save(pipe)
                pipe.execute()
            self.__fragments[fragment_id] = fragment

            return fragment

    def get_fragment_stream(self, fid, until=None):
        return FragmentStream(self.kv, fid).get(until)

    def add_orphaned(self, fid):
        self.__orphaned[fid] = datetime.utcnow() + timedelta(seconds=1000)

    @property
    def orphaned(self):
        return self.__orphaned

    def remove(self, fid):
        fragment = self.__fragments[fid]
        if fid not in self.__orphaned:
            log.info('Discarding fragment: {}'.format(fragment.fid))
            self.add_orphaned(fid)
            self.kv.sadd('{}:orph'.format(self.__fragments_key), fid)
        else:
            now = datetime.utcnow()
            if now > self.__orphaned[fid]:
                log.info('Removing fragment: {}'.format(fragment.fid))
                self.kv.srem(self.__fragments_key, fid)
                self.kv.srem('{}:orph'.format(self.__fragments_key), fid)
                fragment.remove()
                del self.__fragments[fid]

    @staticmethod
    def _daemon():
        futures = {}
        while not stopped.isSet():
            FragmentIndex.daemon_event.clear()
            for index_key in FragmentIndex.instances.keys()[:]:
                index = FragmentIndex.instances[index_key]
                try:
                    with index.lock:
                        for fid in index.fragments.keys()[:]:
                            try:
                                fragment = index.fragments[fid]
                                with fragment.lock:
                                    if fragment.aborted:
                                        index.remove(fragment.fid)
                                    elif not fragment.updated and not fragment.collecting:
                                        if not fragment.newcomer:
                                            index.remove(fragment.fid)
                                        else:
                                            collector = Collector()
                                            collector.planner = index.planner
                                            collector.cache = index.cache
                                            collector.loader = index.loader
                                            collector.force_seed = index.force_seed
                                            log.info('Starting fragment collection: {}'.format(fragment.fid))
                                            try:
                                                futures[fragment.fid] = FragmentIndex.tpool.submit(fragment.populate,
                                                                                                   collector)
                                            except RuntimeError as e:
                                                traceback.print_exc()
                                                log.warn(e.message)
                                    elif fragment.updated and not index.force_seed:
                                        for t, digest in fragment.seed_digests.items():
                                            t_n3 = t.n3(fragment.agp.graph.namespace_manager)
                                            current_digest = index.seed_type_digest(t_n3)
                                            if digest != current_digest:
                                                index.remove(fragment.fid)
                                                break
                            except Exception as e:
                                if index_key in FragmentIndex.instances:
                                    del FragmentIndex.instances[index_key]
                                    # traceback.print_exc()

                    if futures:
                        log.info('Waiting for: {} collections'.format(len(futures)))
                        wait(futures.values())
                        for fragment_id, future in futures.items():
                            exception = future.exception()
                            if exception is not None:
                                log.warn(exception.message)
                                with index.lock:
                                    try:
                                        index.remove(fragment_id)
                                    except Exception, e:
                                        traceback.print_exc()
                                        break

                        futures.clear()
                except AttributeError:
                    pass

            try:
                FragmentIndex.daemon_event.wait(timeout=1)
            except Exception:
                pass


def _map_tp(tp, vars):
    s, p, o = tp
    return TP(vars.get(s, s), vars.get(p, p), vars.get(o, o))


def _apply_filter(v, resource, filters, agp_filters):
    if v in filters:
        for var_f in filters[v]:
            context = QueryContext()
            context[v] = resource
            passing = var_f.expr.eval(context) if hasattr(var_f.expr, 'eval') else bool(
                resource.toPython())
            if not passing:
                return True
    elif v in agp_filters:
        return resource != agp_filters.get(v)
    return False


class Scholar(Collector):
    def __init__(self, id='', **kwargs):
        # type: () -> Scholar
        super(Scholar, self).__init__()
        self.id = id

    def __new__(cls, id='', force_seed=None, **kwargs):
        scholar = super(Scholar, cls).__new__(cls)
        index = FragmentIndex(force_seed=force_seed, id=id, **kwargs)
        scholar.index = index
        return scholar

    @property
    def index(self):
        return self.__index

    @index.setter
    def index(self, i):
        self.__index = i

    @property
    def planner(self):
        return self.index.planner

    @property
    def fountain(self):
        return self.index.planner.fountain

    @property
    def force_seed(self):
        return self.index.force_seed

    @force_seed.setter
    def force_seed(self, s):
        self.index.force_seed = s

    def get(self, agp, filters=None):
        return self.index.get(agp, filters=filters)

    @staticmethod
    def __map(mapping, elm):
        map_terms = mapping.get('vars', None)
        map_literals = mapping.get('filters', None)
        result = elm
        if map_terms:
            result = map_terms.get(result, result)
        if map_literals:
            result = map_literals.get(result, result)

        return result

    def mapped_plan(self, mapping):
        source_plan = mapping['fragment'].plan
        if source_plan:
            mapped_plan = Graph()
            for prefix, uri in source_plan.namespaces():
                mapped_plan.bind(prefix, uri)
            mapped_plan.__iadd__(source_plan)
            v_nodes = list(mapped_plan.subjects(RDF.type, AGORA.Variable))
            for v_node in v_nodes:
                v_source_label = list(mapped_plan.objects(v_node, RDFS.label)).pop()
                mapped_term = self.__map(mapping, Variable(v_source_label))
                if isinstance(mapped_term, Literal):
                    mapped_plan.set((v_node, RDF.type, AGORA.Literal))
                    mapped_plan.set((v_node, AGORA.value, Literal(mapped_term.n3())))
                    mapped_plan.remove((v_node, RDFS.label, None))
                elif isinstance(mapped_term, URIRef):
                    mapped_plan.remove((v_node, None, None))
                    for s, p, _ in mapped_plan.triples((None, None, v_node)):
                        mapped_plan.remove((s, p, v_node))
                        mapped_plan.add((s, p, mapped_term))
                else:
                    mapped_plan.set((v_node, RDFS.label, Literal(mapped_term.n3())))

            return mapped_plan

    def __follow_filter(self, candidates, tp, s, o, trace=None, prev=None):
        def seek_join(tps, f):
            for t in tps:
                for (ss, oo) in filter(lambda x: f(x), candidates[t]):
                    if (ss, t.p, oo) not in trace:
                        yield (t, ss, t.p, oo)
                        trace.add((ss, t.p, oo))
                        for q in self.__follow_filter(candidates, t, ss, oo, trace, prev=tp):
                            yield q

        if trace is None:
            trace = set([])
        trace.add((s, tp.p, o))
        valid_tps = filter(lambda x: x != prev, candidates.keys())
        up_tps = filter(lambda x: x != prev and x.o == tp.s, valid_tps)
        for q in seek_join(up_tps, lambda (ss, oo): oo == s):
            yield q
        down_tps = filter(lambda x: x.s == tp.o, valid_tps)
        for q in seek_join(down_tps, lambda (ss, oo): ss == o):
            yield q
        sib_tps = filter(lambda x: x.s == tp.s and x.o != tp.o, valid_tps)
        for q in seek_join(sib_tps, lambda (ss, oo): ss == s):
            yield q

    def mapped_gen(self, mapping, filters):
        generator = mapping['fragment'].generator
        m_vars = mapping.get('vars', {})
        m_filters = mapping.get('filters', {})
        agp = mapping['fragment'].agp
        if filters == mapping['fragment'].filters:
            filters = {}

        var_filters = {}
        ft = FilterTree()
        in_filter_path = {}
        candidates = {}

        if filters:
            for v in filters:
                var_filters[v] = []
                for f in filters[v]:
                    f = 'SELECT * WHERE { FILTER (%s) }' % f
                    parse = Query.parseString(expandUnicodeEscapes(f), parseAll=True)
                    query = translateQuery(parse)
                    var_filters[v].append(query.algebra.p.p)

            mapped_agp = [_map_tp(tp, m_vars) for tp in agp]
            for tp in mapped_agp:
                ft.add_tp(tp)
                if tp.s in filters or tp.s in m_filters:
                    ft.add_variable(tp.s)
                if tp.o in filters or tp.o in m_filters:
                    ft.add_variable(tp.o)

            ugraph = ft.graph.to_undirected()

            for tp in mapped_agp:
                if tp.s not in in_filter_path:
                    in_filter_path[tp.s] = False
                    for v in ft.variables:
                        try:
                            nx.has_path(ugraph, tp.s, v)
                            in_filter_path[tp.s] = True
                            break
                        except:
                            pass

        for c, s, p, o in generator:
            tp = _map_tp(c, m_vars)

            if not in_filter_path.get(tp.s, False):
                yield tp, s, p, o
            else:
                passing = True
                if _apply_filter(tp.s, s, var_filters, m_filters):
                    ft.filter(s, None, tp.s)
                    passing = False
                if _apply_filter(tp.o, o, var_filters, m_filters):
                    ft.filter(o, None, tp.o)
                    passing = False

                if passing:
                    if tp not in candidates:
                        candidates[tp] = set([])
                    candidates[tp].add((s, o))

        tp_filter_roots = filter(lambda x: x.s in ft.variables or x.o in ft.variables, candidates.keys())
        tp_filter_roots = sorted(tp_filter_roots, key=lambda x: len(ft.graph.successors(x.o)), reverse=True)
        for tp in tp_filter_roots:
            pairs = {}
            for (s, o) in candidates.get(tp, set([])).copy():
                if tp not in pairs:
                    pairs[tp] = set([])
                pairs[tp].add((s, o))
                for ftp, fs, fp, fo in self.__follow_filter(candidates, tp, s, o):
                    if ftp not in pairs:
                        pairs[ftp] = set([])
                    pairs[ftp].add((fs, fo))
                    candidates[ftp].remove((fs, fo))

            if len(candidates) != len(pairs):
                candidates = {}
                break
            else:
                candidates = pairs.copy()

        for tp, pairs in candidates.items():
            for s, o in pairs:
                yield tp, s, tp.p, o

    def get_fragment_generator(self, agp, **kwargs):
        filters = kwargs.get('filters', {})
        mapping = self.index.get(agp, general=True, filters=filters)
        if not mapping:
            # Register fragment
            fragment = self.index.register(agp, filters=filters, follow_cycles=kwargs.get('follow_cycles', True))
            mapping = {'fragment': fragment}

        self.index.notify()
        plan = self.mapped_plan(mapping)
        if plan:
            return {'plan': plan, 'generator': self.mapped_gen(mapping, filters),
                    'prefixes': self.index.planner.fountain.prefixes.items()}
        else:
            raise TypeError('No plan for given agp: {}'.format(agp))

    def shutdown(self):
        self.index.shutdown()



# coding=utf-8
"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import calendar
import logging
import math
import shutil
import zlib
from StringIO import StringIO
from datetime import datetime as dt, timedelta as delta
from threading import Thread, Lock as TLock
from time import sleep

import shortuuid
from concurrent.futures import ThreadPoolExecutor
from rdflib import ConjunctiveGraph
from rdflib.graph import Graph
from redis.lock import Lock

from agora.collector.execution import parse_rdf
from agora.collector.http import http_get, extract_ttl
from agora.engine.utils import stopped
from agora.engine.utils.graph import get_triple_store
from agora.engine.utils.kv import get_kv

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.collector.cache')


class RedisCache(object):
    tpool = ThreadPoolExecutor(max_workers=1)

    def __init__(self, persist_mode=None, key_prefix='', min_cache_time=5, force_cache_time=False,
                 base='store', path='cache', redis_host='localhost', redis_port=6379, redis_db=1, redis_file=None,
                 graph_memory_limit=5000):
        self.__key_prefix = key_prefix
        self.__cache_key = '{}:cache'.format(key_prefix)
        self.__persist_mode = persist_mode
        self.__min_cache_time = min_cache_time
        self.__force_cache_time = force_cache_time
        self.__base_path = base
        self._r = get_kv(persist_mode, redis_host, redis_port, redis_db, redis_file, base=base, path=path)
        self.__lock = Lock(self._r, key_prefix)
        self.__mlock = TLock()
        self.__graph_memory_limit = graph_memory_limit
        self.__memory_graphs = {}
        self.__memory_order = []

        self.__resources_ts = {}

        for lock_key in self._r.keys('{}:l*'.format(self.__key_prefix)):
            self._r.delete(lock_key)

        self._r.delete(key_prefix)

        self.__enabled = True
        self.__purge_th = Thread(target=self.__purge)
        self.__purge_th.daemon = True
        self.__purge_th.start()

    @property
    def r(self):
        return self._r

    @r.setter
    def r(self, r):
        self._r = r

    @property
    def min_cache_time(self):
        return self.__min_cache_time

    @min_cache_time.setter
    def min_cache_time(self, t):
        self.__min_cache_time = t

    @property
    def resource_cache(self):
        return self.__resource_cache

    @resource_cache.setter
    def resource_cache(self, rc):
        self.__resource_cache = rc

    @property
    def persist_mode(self):
        return self.__persist_mode

    @property
    def base_path(self):
        return self.__base_path

    def __clean(self, name):
        shutil.rmtree('{}/{}'.format(self.__base_path, name))

    def uri_lock(self, uri):
        with self.__lock:
            key = '{}:l:'.format(self.__key_prefix) + uri
            key = (key[:250]) if len(key) > 250 else key
            return Lock(self._r, key)

    def __purge(self):
        gids_key = '{}:gids'.format(self.__cache_key)
        while self.__enabled and not stopped.isSet():
            try:
                gids = self._r.hkeys(gids_key)
                obsolete = filter(
                    lambda x: not self._r.exists('{}:{}'.format(self.__cache_key, self._r.hget(gids_key, x))),
                    gids)

                if obsolete:
                    with self._r.pipeline(transaction=True) as p:
                        p.multi()
                        log.debug('Removing {} resouces from cache...'.format(len(obsolete)))
                        for uri in obsolete:
                            lock = self.uri_lock(uri)
                            with lock:
                                try:
                                    self._r.hdel(gids_key, uri)
                                    self.__forget(uri)
                                except Exception:
                                    # traceback.print_exc()
                                    log.error('Purging resource {}'.format(uri))
                                p.execute()
            except Exception, e:
                if not stopped.isSet():
                    log.error(e.message)
                self.__enabled = False
            sleep(10)

    def __memoize(self, key, graph):
        with self.__mlock:
            if len(self.__memory_order) >= self.__graph_memory_limit:
                old_key = self.__memory_order.pop(0)
                if old_key in self.__memory_graphs:
                    del self.__memory_graphs[old_key]

            self.__memory_order.append(key)
            self.__memory_graphs[key] = graph

    def __recall(self, key):
        with self.__mlock:
            return self.__memory_graphs[key]

    def __forget(self, key):
        with self.__mlock:
            try:
                if key in self.__memory_graphs:
                    del self.__memory_graphs[key]
                    self.__memory_order.remove(key)
            except:
                pass

    def release_locks(self):
        with self.__lock:
            lock_keys = self._r.keys('{}:l:*'.format(self.__key_prefix))
            for k in lock_keys:
                self._r.delete(k)

    def create(self, conjunctive=False, gid=None, loader=None, format=None):
        if conjunctive:
            uuid = shortuuid.uuid()
            g = get_triple_store(self.__persist_mode, base=self.__base_path, path=uuid)
            return g
        else:
            p = self._r.pipeline(transaction=True)
            p.multi()

            g = Graph(identifier=gid)

            lock = self.uri_lock(gid)
            with lock:
                uuid = self._r.hget('{}:gids'.format(self.__cache_key), gid)
                if not uuid:
                    uuid = shortuuid.uuid()
                    p.hset('{}:gids'.format(self.__cache_key), gid, uuid)

                gid_key = '{}:{}'.format(self.__cache_key, uuid)

                ttl_ts = self._r.hget(gid_key, 'ttl')
                if ttl_ts is not None:
                    ttl_dt = dt.utcfromtimestamp(int(ttl_ts))
                    now = dt.utcnow()
                    if ttl_dt > now:
                        try:
                            g = self.__recall(gid)
                        except KeyError:
                            source_z = self._r.hget(gid_key, 'data')
                            source = zlib.decompress(source_z)
                            g.parse(StringIO(source), format=format)
                            self.__memoize(gid, g)

                        ttl = math.ceil((ttl_dt - dt.utcnow()).total_seconds())
                        return g, math.ceil(ttl)

                log.debug('Caching {}'.format(gid))
                response = loader(gid, format)
                if response is None and loader != http_get:
                    response = http_get(gid, format)

                if isinstance(response, bool):
                    return response

                ttl = self.__min_cache_time
                source, headers = response
                if not isinstance(source, Graph) and not isinstance(source, ConjunctiveGraph):
                    parse_rdf(g, source, format, headers)
                    data = g.serialize(format='turtle')
                else:
                    data = source.serialize(format='turtle')
                    for prefix, ns in source.namespaces():
                        g.bind(prefix, ns)
                    g.__iadd__(source)

                self.__memoize(gid, g)

                if not self.__force_cache_time:
                    ttl = extract_ttl(headers) or ttl

                p.hset(gid_key, 'data', zlib.compress(data))
                ttl_ts = calendar.timegm((dt.utcnow() + delta(seconds=ttl)).timetuple())
                p.hset(gid_key, 'ttl', ttl_ts)
                p.expire(gid_key, ttl)
                p.execute()
            return g, int(ttl)

    def release(self, g):
        if isinstance(g, ConjunctiveGraph):
            if self.__persist_mode:
                RedisCache.tpool.submit(self.__clean, g.identifier.toPython())
            else:
                g.remove((None, None, None))

    def expire(self, gid):
        lock = self.uri_lock(gid)
        with lock:
            uuid = self._r.hget('{}:gids'.format(self.__cache_key), gid)
            with self._r.pipeline(transaction=True) as p:
                if not uuid:
                    uuid = shortuuid.uuid()
                    p.hset('{}:gids'.format(self.__cache_key), gid, uuid)

                gid_key = '{}:{}'.format(self.__cache_key, uuid)
                p.delete(gid_key)
                p.execute()

    def get_matching_uris(self, part):
        return filter(lambda gid: part in gid, self._r.hkeys('{}:gids'.format(self.__cache_key)))

    def close(self):
        self.__enabled = False



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import calendar
import logging
import traceback
from datetime import datetime as dt

import redis
from concurrent.futures import wait

from agora.engine.fountain.schema import Schema
from agora.engine.utils.cache import cached
from agora.engine.utils.kv import close_kv

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.fountain.index')


def __get_by_pattern(r, pattern, func):
    # type: (redis.StrictRedis, str, callable) -> list

    def get_all():
        for k in pkeys:
            yield func(k)

    pkeys = r.keys(pattern)
    return list(get_all())


def __remove_from_sets(r, values, *args):
    # type: (redis.StrictRedis, iter, iter) -> None

    for pattern in args:
        keys = r.keys(pattern)
        for dk in keys:
            key_parts = dk.split(':')
            ef_values = values
            if len(key_parts) > 1:
                ef_values = filter(lambda x: x.split(':')[0] != key_parts[1], values)
            if len(ef_values):
                r.srem(dk, *ef_values)


def __get_vocab_set(r, pattern, vid=None):
    # type: (redis.StrictRedis, str, str) -> list

    if vid is not None:
        pattern = pattern.replace(':*:', ':%s:' % vid)
    all_sets = map(lambda x: r.smembers(x), r.keys(pattern))
    return list(reduce(set.union, all_sets, set([])))


def __extract_type(schema, r, t, vid):
    # type: (Schema, redis.StrictRedis, str, str) -> None

    log.debug('Extracting type {} from the {} vocabulary...'.format(t, vid))
    with r.pipeline() as pipe:
        pipe.multi()
        pipe.sadd('vocabs:{}:types'.format(vid), t)
        for s in schema.get_supertypes(t):
            pipe.sadd('vocabs:{}:types:{}:super'.format(vid, t), s)
        for s in schema.get_subtypes(t):
            pipe.sadd('vocabs:{}:types:{}:sub'.format(vid, t), s)
        for s in schema.get_type_properties(t):
            pipe.sadd('vocabs:{}:types:{}:props'.format(vid, t), s)
        for s in schema.get_type_specific_properties(t):
            pipe.sadd('vocabs:{}:types:{}:s-props'.format(vid, t), s)
        for s in schema.get_type_references(t):
            pipe.sadd('vocabs:{}:types:{}:refs'.format(vid, t), s)
        for s in schema.get_type_specific_references(t):
            pipe.sadd('vocabs:{}:types:{}:s-refs'.format(vid, t), s)
        pipe.execute()


def __extract_property(schema, r, p, vid):
    # type: (Schema, redis.StrictRedis, str, str) -> None

    def p_type():
        if schema.is_object_property(p):
            return 'object'
        else:
            return 'data'

    try:
        log.debug('Extracting property {} from the {} vocabulary...'.format(p, vid))
        with r.pipeline() as pipe:
            pipe.multi()
            pipe.sadd('vocabs:{}:properties'.format(vid), p)
            pipe.hset('vocabs:{}:properties:{}'.format(vid, p), 'uri', p)
            for dc in list(schema.get_property_domain(p)):
                pipe.sadd('vocabs:{}:properties:{}:_domain'.format(vid, p), dc)
            for dc in list(schema.get_property_range(p)):
                pipe.sadd('vocabs:{}:properties:{}:_range'.format(vid, p), dc)
            for dc in list(schema.get_property_inverses(p)):
                pipe.sadd('vocabs:{}:properties:{}:_inverse'.format(vid, p), dc)

            for cn in list(schema.get_property_constraints(p)):
                pipe.sadd('vocabs:{}:properties:{}:_cons'.format(vid, p), cn)

            pipe.set('vocabs:{}:properties:{}:_type'.format(vid, p), p_type())
            pipe.execute()
    except TypeError, e:
        traceback.print_exc()


def __extract_types(schema, r, vid, trace=None):
    # type: (Schema, redis.StrictRedis, str, set) -> iter
    types = schema.get_types(context=vid)

    other_vocabs = filter(lambda x: x != vid, schema.contexts)
    dependent_types = set([])
    dependent_props = set([])
    for ovid in other_vocabs:
        o_types = [t for t in _get_types(r, ovid) if t not in types]
        for oty in o_types:
            otype = _get_type(r, oty)
            if set.intersection(types, otype.get('super')) or set.intersection(types, otype.get('sub')):
                dependent_types.add((ovid, oty))
        o_props = [t for t in _get_properties(r, ovid)]
        for op in o_props:
            oprop = _get_property(r, op)
            if set.intersection(types, oprop.get('domain')) or set.intersection(types, oprop.get('range')):
                dependent_props.add((ovid, op))

    types = set.union(set([(vid, t) for t in types]), dependent_types)
    futures = []
    if trace is None:
        trace = set([])
    for v, t in types:
        if t not in trace:
            __extract_type(schema, r, t, v)
            trace.add(t)
    for v, p in dependent_props:
        if p not in trace:
            __extract_property(schema, r, p, v)
            trace.add(p)
    return types, futures


def __extract_properties(schema, r, vid, trace=None):
    # type: (Schema, redis.StrictRedis, str, set) -> iter

    properties = schema.get_properties(context=vid)

    other_vocabs = filter(lambda x: x != vid, schema.contexts)
    dependent_types = set([])
    for ovid in other_vocabs:
        o_types = [t for t in _get_types(r, ovid)]
        for oty in o_types:
            o_type = _get_type(r, oty)
            if set.intersection(properties, o_type.get('refs')) or set.intersection(properties,
                                                                                    o_type.get('properties')):
                dependent_types.add((ovid, oty))

    futures = []
    if trace is None:
        trace = set([])
    for p in properties:
        if p not in trace:
            __extract_property(schema, r, p, vid)
            trace.add(p)
    for v, ty in dependent_types:
        if ty not in trace:
            __extract_type(schema, r, ty, v)
            trace.add(ty)
    return properties, futures


def _delete_vocabulary(r, vid):
    # type: (redis.StrictRedis, str) -> None

    v_types = _get_types(r, vid)
    if len(v_types):
        __remove_from_sets(r, v_types, '*:_domain', '*:_range', '*:sub', '*:super')
    v_props = _get_properties(r, vid)
    if len(v_props):
        __remove_from_sets(r, v_props, '*:refs', '*:props')

    v_keys = r.keys('vocabs:{}:*'.format(vid))
    if len(v_keys):
        r.delete(*v_keys)


def _extract_vocabulary(schema, r, vid):
    # type: (Schema, redis.StrictRedis, str) -> (iter, iter)

    log.info('Extracting vocabulary {}...'.format(vid))
    _delete_vocabulary(r, vid)
    start_time = dt.now()
    r.set('ts', calendar.timegm(start_time.timetuple()))
    extracted = set([])
    types, t_futures = __extract_types(schema, r, vid, trace=extracted)
    properties, p_futures = __extract_properties(schema, r, vid, trace=extracted)
    wait(p_futures + t_futures)
    log.info('Done (in {}ms)'.format((dt.now() - start_time).total_seconds() * 1000))
    return types, properties


def _get_types(r, vid=None):
    # type: (redis.StrictRedis, str) -> iter

    def shared_type(t):
        return any(filter(lambda k: r.sismember(k, t), all_type_keys))

    vid_types = __get_vocab_set(r, 'vocabs:*:types', vid)

    if vid is not None:
        all_type_keys = filter(lambda k: k != 'vocabs:{}:types'.format(vid), r.keys("vocabs:*:types"))
        vid_types = filter(lambda t: not shared_type(t), vid_types)

    return vid_types


def _get_properties(r, vid=None):
    # type: (redis.StrictRedis, str) -> iter

    def shared_property(t):
        return any(filter(lambda k: r.sismember(k, t), all_prop_keys))

    vid_props = __get_vocab_set(r, 'vocabs:*:properties', vid)

    if vid is not None:
        all_prop_keys = filter(lambda k: k != 'vocabs:{}:properties'.format(vid), r.keys("vocabs:*:properties"))
        vid_props = filter(lambda t: not shared_property(t), vid_props)

    return vid_props


def _get_property(r, prop):
    # type: (redis.StrictRedis, str) -> dict

    def get_inverse_domain(ip):
        return reduce(set.union, __get_by_pattern(r, '*:properties:{}:_domain'.format(ip), r.smembers), set([]))

    def get_inverse_range(ip):
        return reduce(set.union, __get_by_pattern(r, '*:properties:{}:_range'.format(ip), r.smembers), set([]))

    all_prop_keys = r.keys('*:properties')
    if not filter(lambda k: r.sismember(k, prop), all_prop_keys):
        raise TypeError('Unknown property')

    domain = reduce(set.union, __get_by_pattern(r, '*:properties:{}:_domain'.format(prop), r.smembers), set([]))
    rang = reduce(set.union, __get_by_pattern(r, '*:properties:{}:_range'.format(prop), r.smembers), set([]))
    inv = reduce(set.union, __get_by_pattern(r, '*:properties:{}:_inverse'.format(prop), r.smembers), set([]))
    cons = reduce(set.union, __get_by_pattern(r, '*:properties:{}:_cons'.format(prop), r.smembers), set([]))
    cons = map(lambda x: eval(x), cons)
    if len(inv):
        inverse_dr = [(get_inverse_domain(i), get_inverse_range(i)) for i in inv]
        for dom, ra in inverse_dr:
            domain.update(ra)
            rang.update(dom)

    ty = __get_by_pattern(r, '*:properties:{}:_type'.format(prop), r.get)
    try:
        ty = ty.pop()
    except IndexError:
        ty = 'object'

    return {'domain': list(domain),
            'range': list(rang),
            'inverse': list(inv),
            'constraints': dict(cons),
            'type': ty}


def _is_property(r, prop):
    # type: (redis.StrictRedis, str) -> bool
    all_prop_keys = r.keys('*:properties')
    return bool(filter(lambda k: r.sismember(k, prop), all_prop_keys))


def _is_type(r, ty):
    # type: (redis.StrictRedis, str) -> bool
    all_type_keys = r.keys('*:types')
    return bool(filter(lambda k: r.sismember(k, ty), all_type_keys))


def _get_type(r, ty):
    # type: (redis.StrictRedis, str) -> dict
    all_type_keys = r.keys('*:types')
    if not filter(lambda k: r.sismember(k, ty), all_type_keys):
        raise TypeError('Unknown type: {}'.format(ty))

    super_types = reduce(set.union, __get_by_pattern(r, '*:types:{}:super'.format(ty), r.smembers), set([]))
    sub_types = reduce(set.union, __get_by_pattern(r, '*:types:{}:sub'.format(ty), r.smembers), set([]))
    type_props = reduce(set.union, __get_by_pattern(r, '*:types:{}:props'.format(ty), r.smembers), set([]))
    type_s_props = reduce(set.union, __get_by_pattern(r, '*:types:{}:s-props'.format(ty), r.smembers), set([]))
    type_refs = reduce(set.union, __get_by_pattern(r, '*:types:{}:refs'.format(ty), r.smembers), set([]))
    type_s_refs = reduce(set.union, __get_by_pattern(r, '*:types:{}:s-refs'.format(ty), r.smembers), set([]))

    return {'super': list(super_types),
            'sub': list(sub_types),
            'properties': list(type_props),
            'spec_properties': list(type_s_props),
            'refs': list(type_refs),
            'spec_refs': list(type_s_refs)}


def _cached(f):
    # type: (callable) -> callable
    def wrap(self=None, *args, **kwargs):
        if self.schema.cache is not None:
            return cached(self.schema.cache, ref_ts=self.ts)(f)(self, *args, **kwargs)
        else:
            return f(self, *args, **kwargs)

    return wrap


class Index(object):
    def __init__(self):
        self.__schema = None
        self.__r = None

    @property
    def schema(self):
        # type: () -> Schema
        return self.__schema

    @property
    def ts(self):
        # type: () -> int
        ts = self.__r.get('ts')
        ts = 0 if ts is None else int(ts)
        return ts

    @schema.setter
    def schema(self, schema):
        # type: (Schema) -> None
        self.__schema = schema

    @property
    def r(self):
        # type: () -> redis.StrictRedis
        return self.__r

    @r.setter
    def r(self, r):
        # type: (redis.StrictRedis) -> None
        self.__r = r

    @property
    def types(self):
        return _get_types(self.r)

    @property
    def properties(self):
        return _get_properties(self.r)

    @_cached
    def get_type(self, t):
        # type: (str) -> dict
        return _get_type(self.r, t)

    @_cached
    def get_property(self, p):
        # type: (str) -> dict
        return _get_property(self.r, p)

    @_cached
    def is_type(self, t):
        return _is_type(self.r, t)

    @_cached
    def is_property(self, p):
        return _is_property(self.r, p)

    def index_vocabulary(self, vid):
        return _extract_vocabulary(self.__schema, self.__r, vid)

    def delete_vocabulary(self, vid):
        return _delete_vocabulary(self.__r, vid)

    def close(self):
        close_kv(self.r)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import base64
import collections
import logging

import redis
from rfc3987 import parse

from agora.engine.fountain.index import Index
from agora.engine.utils.cache import Cache

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.fountain.seed')


class SeedManager(object):
    def __init__(self):
        # type: () -> SeedManager
        self.__index = None
        self.__cache = Cache()

    @property
    def index(self):
        return self.__index

    @index.setter
    def index(self, i):
        self.__index = i
        self.__cache.watch(self.__index.schema.cache)

    def add_seed(self, uri, type):
        # type: (str, str) -> str
        sid = _add_seed(self.__index.r, uri, type)
        self.__cache.clear()
        return sid

    def delete_seed(self, sid):
        # type: (str) -> None
        _delete_seed(self.__index.r, sid)
        self.__cache.clear()

    def get_seed(self, sid):
        # type: (str) -> dict
        return _get_seed(self.__index.r, sid)

    def delete_type_seeds(self, type):
        # type: (str) -> None
        _delete_type_seeds(self.__index.r, type)
        self.__cache.clear()

    @property
    def seeds(self):
        # type: () -> dict
        return _get_seeds(self.__index.r)

    def get_type_seeds(self, type):
        # type: (str) -> iter
        return _get_type_seeds(self.__index, type)

    def validate(self):
        seed_dict = self.seeds
        types = self.__index.types
        obsolete_types = set.difference(set(seed_dict.keys()), set(types))
        for t in obsolete_types:
            self.delete_type_seeds(t)


class DuplicateSeedError(Exception):
    pass


class InvalidSeedError(Exception):
    pass


def _add_seed(r, uri, ty):
    # type: (redis.StrictRedis, str, str) -> str
    parse(uri, rule='URI')
    type_found = False
    type_keys = r.keys('*:types')
    for tk in type_keys:
        if r.sismember(tk, ty):
            type_found = True
            encoded_uri = base64.b64encode(uri)
            if r.sismember('seeds:{}'.format(ty), encoded_uri):
                raise DuplicateSeedError('{} is already registered as a seed of type {}'.format(uri, ty))
            r.sadd('seeds:{}'.format(ty), base64.b64encode(uri))
            break

    if not type_found:
        raise TypeError("{} is not a valid type".format(ty))

    return base64.b64encode('{}|{}'.format(ty, uri))


def _get_seed(r, sid):
    # type: (redis.StrictRedis, str) -> dict
    try:
        ty, uri = base64.b64decode(sid).split('|')
        if r.sismember('seeds:{}'.format(ty), base64.b64encode(uri)):
            return {'type': ty, 'uri': uri}
    except (ValueError, TypeError) as e:
        raise InvalidSeedError(e.message)

    raise InvalidSeedError(sid)


def _delete_seed(r, sid):
    # type: (redis.StrictRedis, str) -> None
    try:
        ty, uri = base64.b64decode(sid).split('|')
        set_key = 'seeds:{}'.format(ty)
        encoded_uri = base64.b64encode(uri)
        if not r.srem(set_key, encoded_uri):
            raise InvalidSeedError(sid)
    except TypeError as e:
        raise InvalidSeedError(e.message)


def _delete_type_seeds(r, ty):
    # type: (redis.StrictRedis, str) -> None
    r.delete('seeds:{}'.format(ty))


def _get_seeds(r):
    # type: (redis.StrictRedis) -> dict
    def iterator():
        seed_types = r.keys('seeds:*')
        for st in seed_types:
            ty = st.replace('seeds:', '')
            for seed in list(r.smembers(st)):
                yield ty, base64.b64decode(seed)

    result_dict = collections.defaultdict(list)
    for t, uri in iterator():
        result_dict[t].append({"uri": uri, "id": base64.b64encode('{}|{}'.format(t, uri))})
    return result_dict


def _get_type_seeds(index, ty):
    # type: (Index) -> iter
    try:
        t_dict = index.get_type(ty)
        all_seeds = set([])
        for t in t_dict['sub'] + [ty]:
            all_seeds.update([base64.b64decode(seed) for seed in list(index.r.smembers('seeds:{}'.format(t)))])
        return list(all_seeds)
    except TypeError:
        # Check if it is a property...and return an empty list
        try:
            index.get_property(ty)
            return []
        except TypeError:
            raise TypeError(ty)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import re
from abc import abstractmethod

from rdflib import BNode
from rdflib import Literal
from rdflib import URIRef
from shortuuid import uuid

from agora.collector.cache import RedisCache
from agora.collector.execution import PlanExecutor
from agora.engine.plan.agp import AGP
from agora.engine.plan.graph import AGORA
from agora.engine.utils import Wrapper

__author__ = "Fernando Serena"

log = logging.getLogger('agora.collector')


class AbstractCollector(object):
    @property
    @abstractmethod
    def prefixes(self):
        raise NotImplementedError

    @abstractmethod
    def get_fragment_generator(self, agp, **kwargs):
        # type: (AGP, dict) -> iter
        raise NotImplementedError


class Collector(AbstractCollector):
    def __init__(self):
        # type: () -> Collector
        self.__cache = None
        self.__planner = None
        self.__loader = None
        self.__force_seed = None
        self.__fountain = None

    @property
    def loader(self):
        return self.__loader

    @loader.setter
    def loader(self, l):
        self.__loader = l

    @property
    def force_seed(self):
        return self.__force_seed

    @force_seed.setter
    def force_seed(self, s):
        self.__force_seed = s

    @property
    def planner(self):
        return self.__planner

    @planner.setter
    def planner(self, p):
        self.__planner = p

    @property
    def fountain(self):
        if not isinstance(self.__fountain, Wrapper):
            self.__fountain = Wrapper(self.__planner.fountain)
        return self.__fountain

    @property
    def cache(self):
        return self.__cache

    @cache.setter
    def cache(self, c):
        self.__cache = c

    def get_fragment_generator(self, agp, **kwargs):
        # type: (AGP) -> dict

        force_seed_tuples = []
        if self.__force_seed:
            for ty in self.__force_seed.keys():
                force_seed_tuples.append((URIRef('http://{}'.format(uuid())), ty))
        plan = self.__planner.make_plan(agp, force_seed=force_seed_tuples)

        seed_triples_to_remove = set()
        for s, ty in force_seed_tuples:
            trees = list(plan.subjects(predicate=AGORA.hasSeed, object=s))
            for t in trees:
                seed_triples_to_remove.add((t, AGORA.hasSeed, s))
                for actual_s in self.__force_seed[ty]:
                    plan.add((t, AGORA.hasSeed, actual_s))
        for t in seed_triples_to_remove:
            plan.remove(t)

        executor = PlanExecutor(plan)

        def with_context_ttl():
            return executor.ttl

        def with_context_derefs():
            return executor.n_derefs

        fragment_dict = executor.get_fragment_generator(cache=self.cache, loader=self.__loader, **kwargs)
        fragment_dict['ttl'] = with_context_ttl
        fragment_dict['n_derefs'] = with_context_derefs
        return fragment_dict

    @property
    def prefixes(self):
        return self.fountain.prefixes


def triplify(x):
    def __extract_lang(v):
        def __lang_tag_match(strg, search=re.compile(r'[^a-z]').search):
            return not bool(search(strg))

        if '@' in v:
            try:
                (v_aux, lang) = tuple(v.split('@'))
                (v, lang) = (v_aux, lang) if __lang_tag_match(lang) else (v, None)
            except ValueError:
                lang = None
        else:
            lang = None
        return v, lang

    def __term(elm):
        if elm.startswith('<'):
            return URIRef(elm.lstrip('<').rstrip('>'))
        elif '^^' in elm:
            (value, ty) = tuple(elm.split('^^'))
            return Literal(value.replace('"', ''), datatype=URIRef(ty.lstrip('<').rstrip('>')))
        elif elm.startswith('_:'):
            return BNode(elm.replace('_:', ''))
        else:
            (elm, lang) = __extract_lang(elm)
            elm = elm.replace('"', '')
            if lang is not None:
                return Literal(elm, lang=lang)
            else:
                return Literal(elm)

    c, s, p, o = eval(x)
    return c, __term(s), __term(p), __term(o)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging

from rdflib import URIRef, BNode, Graph
from rdflib.namespace import RDFS, OWL, NamespaceManager

from agora.engine.utils.cache import Cache, ContextGraph, cached

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.fountain.path')


def __flat_slice(iterable):
    # type: (iter) -> set
    lst = filter(lambda x: x, list(iterable))
    for i, _ in enumerate(lst):
        while hasattr(lst[i], "__iter__") and not isinstance(lst[i], basestring):
            lst[i:i + 1] = lst[i]
    return set(filter(lambda x: x is not None, lst))


def __q_name(ns, term):
    # type: (NamespaceManager, any) -> any
    n3_method = getattr(term, "n3", None)
    if callable(n3_method):
        return term.n3(ns)
    return term


def __query(graph, q):
    # type: (Graph, str) -> set
    graph.namespace_manager.bind('rdfs', RDFS, replace=True, override=True)
    graph.namespace_manager.bind('owl', OWL, replace=True, override=True)
    result = graph.query(q)
    return set([__q_name(graph.namespace_manager, x) for x in __flat_slice(result)])


def __extend_prefixed(ns, pu):
    # type: (dict, str) -> URIRef

    parts = pu.split(':')
    if len(parts) == 1:
        parts = ('', parts[0])
    try:
        return URIRef(ns[parts[0]] + parts[1])
    except KeyError:
        return BNode(pu)


def __extend_with(f, graph, *args):
    # type: (callable, Graph, iter) -> iter
    args = __flat_slice(args)
    extension = __flat_slice([f(graph, t) for t in args])
    return set.union(args, extension)


def _contexts(graph):
    # type: (ContextGraph) -> list
    return [str(x.identifier) for x in graph.contexts()]


def _update_context(graph, vid, g):
    # type: (ContextGraph, str, Graph) -> None
    context = graph.get_context(vid)
    if isinstance(context, SchemaGraph):
        context = context.g
    graph.remove_context(context)
    _add_context(graph, vid, g)


def _remove_context(graph, vid):
    # type: (ContextGraph, str) -> None
    context = graph.get_context(vid)
    if isinstance(context, SchemaGraph):
        context = context.g
    graph.remove_context(context)


def _get_context(graph, vid):
    # type: (ContextGraph, str) -> Graph
    return graph.get_context(vid)


def _add_context(graph, vid, g):
    # type: (ContextGraph, str, Graph) -> None
    def match_ns(term):
        filter_ns = [ns for ns in rev_ns if ns in term]
        if filter_ns:
            ns = filter_ns.pop()
            if rev_ns[ns]:
                vid_context.bind(rev_ns[ns], ns)
            del rev_ns[ns]

    rev_ns = {ns: prefix for prefix, ns in g.namespaces() if not prefix.startswith('ns')}
    vid_context = graph.get_context(vid)
    for s, p, o in g.triples((None, None, None)):
        if o != OWL.Ontology:
            match_ns(s)
            match_ns(p)
            match_ns(o)
        vid_context.add((s, p, o))


def _prefixes(graph):
    # type: (Graph) -> dict
    return dict(graph.namespaces())


def _get_types(graph):
    # type: (Graph) -> set
    return __query(graph,
                   """SELECT DISTINCT ?c WHERE {
                        {
                            ?p a owl:ObjectProperty .
                            {
                                { ?p rdfs:range ?c }
                                UNION
                                { ?p rdfs:domain ?c }
                            }
                        }
                        UNION
                        {
                            ?p a owl:DatatypeProperty .
                            ?p rdfs:domain ?c .
                        }
                        UNION
                        { ?c a owl:Class }
                        UNION
                        { ?c a rdfs:Class }
                        UNION
                        { [] rdfs:subClassOf ?c }
                        UNION
                        { ?c rdfs:subClassOf [] }
                        UNION
                        {
                            ?r a owl:Restriction ;
                               owl:onProperty ?p .
                            {
                                ?p a owl:ObjectProperty .
                                { ?r owl:allValuesFrom ?c }
                                UNION
                                { ?r owl:someValuesFrom ?c }
                            }
                            UNION
                            { ?r owl:onClass ?c }
                        }
                        FILTER(isURI(?c))
                      }""")


def _get_properties(graph):
    # type: (Graph) -> set
    return __query(graph, """SELECT DISTINCT ?p WHERE {
                                { ?p a rdf:Property }
                                UNION
                                { ?p a owl:ObjectProperty }
                                UNION
                                { ?p a owl:DatatypeProperty }
                                UNION
                                {
                                    [] a owl:Restriction ;
                                       owl:onProperty ?p .
                                }
                                FILTER(isURI(?p))
                              }""")


def _is_object_property(graph, prop):
    # type: (Graph, str) -> bool
    evidence = __query(graph, """ASK {
                                    { %s a owl:ObjectProperty }
                                    UNION
                                    {
                                        ?r owl:onProperty %s .
                                        {
                                            { ?c a owl:Class }
                                            UNION
                                            { ?c rdfs:subClassOf ?r }
                                        }
                                        {
                                            {
                                               ?r owl:onClass ?c .
                                            }
                                            UNION
                                            {
                                                ?r owl:someValuesFrom ?c .
                                            }
                                            UNION
                                            {
                                                ?r owl:allValuesFrom ?c .
                                            }
                                        }
                                    }
                                   }""" % (prop, prop))

    return False if not evidence else bool(evidence.pop())


def _get_property_domain(graph, prop):
    # type: (Graph, str) -> set
    all_property_domains = graph.query("""
                          PREFIX owl: <http://www.w3.org/2002/07/owl#>
                          PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                          SELECT DISTINCT ?p ?c WHERE {
                             { ?p rdfs:domain ?c }
                             UNION
                             { ?c rdfs:subClassOf [ owl:onProperty ?p ] }
                             FILTER (isURI(?p) && isURI(?c))
                           }""")

    dom = map(lambda x: __q_name(graph.namespace_manager, x.c),
              filter(lambda x: __q_name(graph.namespace_manager, x.p) == prop, all_property_domains))
    return __extend_with(_get_subtypes, graph, dom)


def _get_property_range(graph, prop):
    # type: (Graph, str) -> set
    all_property_ranges = graph.query("""
                                        PREFIX owl: <http://www.w3.org/2002/07/owl#>
                                        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                                        SELECT DISTINCT ?p ?r WHERE {
                                              {?p rdfs:range ?r}
                                              UNION
                                              {
                                                    ?d owl:onProperty ?p.
                                                    { ?d owl:allValuesFrom ?r }
                                                    UNION
                                                    { ?d owl:someValuesFrom ?r }
                                                    UNION
                                                    { ?d owl:onClass ?r }
                                                    UNION
                                                    { ?d owl:onDataRange ?r }
                                              }
                                              FILTER(isURI(?p) && isURI(?r))
                                            }""")

    rang = map(lambda x: __q_name(graph.namespace_manager, x.r),
               filter(lambda x: __q_name(graph.namespace_manager, x.p) == prop, all_property_ranges))
    return __extend_with(_get_subtypes, graph, rang)


def _get_property_inverses(graph, prop):
    # type: (Graph, str) -> set
    return __query(graph, """SELECT DISTINCT ?i WHERE {
                                 {%s owl:inverseOf ?i}
                                 UNION
                                 {?i owl:inverseOf %s}
                                 FILTER(isURI(?i))
                               }""" % (prop, prop))


def _get_property_constraints(graph, prop):
    # type: (Graph, str) -> set
    all_property_domains = graph.query("""
                               PREFIX owl: <http://www.w3.org/2002/07/owl#>
                               PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                               SELECT DISTINCT ?p ?c WHERE {
                                 { ?p rdfs:domain ?c }
                                 UNION
                                 { ?c rdfs:subClassOf [ owl:onProperty ?p ] }
                                 FILTER (isURI(?p) && isURI(?c))
                               }""")

    dom = map(lambda x: __q_name(graph.namespace_manager, x.c),
              filter(lambda x: __q_name(graph.namespace_manager, x.p) == prop, all_property_domains))
    dom_supertypes = [_get_supertypes(graph, d) for d in dom]
    for d, s in zip(dom, dom_supertypes):
        if set.intersection(set(s), set(dom)):
            cons_range = __query(graph, """SELECT DISTINCT ?r WHERE {
                                              {
                                                    %s rdfs:subClassOf ?d .
                                                    ?d owl:onProperty %s .
                                                    { ?d owl:allValuesFrom ?r }
                                                    UNION
                                                    { ?d owl:someValuesFrom ?r }
                                                    UNION
                                                    { ?d owl:onClass ?r }
                                                    UNION
                                                    { ?d owl:onDataRange ?r }
                                              }
                                              FILTER(isURI(?r))
                                            }""" % (d, prop))
            cons_range = __extend_with(_get_subtypes, graph, cons_range)
            if cons_range:
                try:
                    is_strong_constraint = __query(graph, """ASK {                                                          
                                                                %s rdfs:subClassOf ?d .
                                                                ?d owl:onProperty %s .
                                                                ?d owl:allValuesFrom []                                                               
                                                            }""" % (d, prop)).pop()
                except KeyError:
                    is_strong_constraint = True

                if is_strong_constraint:
                    yield (d, list(cons_range))


def _get_supertypes(graph, ty):
    # type: (Graph, str) -> set
    res = map(lambda x: __q_name(graph.namespace_manager, x), filter(lambda y: isinstance(y, URIRef),
                                                                     graph.transitive_objects(
                                                                         __extend_prefixed(_prefixes(graph), ty),
                                                                         RDFS.subClassOf)))
    return set(filter(lambda x: str(x) != ty, res))


def _get_subtypes(graph, ty):
    # type: (Graph, str) -> set
    res = map(lambda x: __q_name(graph.namespace_manager, x), filter(lambda y: isinstance(y, URIRef),
                                                                     graph.transitive_subjects(RDFS.subClassOf,
                                                                                               __extend_prefixed(
                                                                                                   _prefixes(graph),
                                                                                                   ty))))

    return set(filter(lambda x: str(x) != ty, res))


def _get_type_properties(graph, ty):
    # type: (Graph, str) -> set
    all_class_props = graph.query("""
                                    PREFIX owl: <http://www.w3.org/2002/07/owl#>
                                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                                    SELECT DISTINCT ?c ?p WHERE {
                                        {?c rdfs:subClassOf [ owl:onProperty ?p ] }
                                        UNION
                                        {?p rdfs:domain ?c}
                                        FILTER (isURI(?p) && isURI(?c))
                                      }""")

    all_types = __extend_with(_get_supertypes, graph, ty)
    return set([__q_name(graph.namespace_manager, r.p) for r in all_class_props if
                __q_name(graph.namespace_manager, r.c) in all_types])


def _get_type_specific_properties(graph, ty):
    # type: (Graph, str) -> set
    all_class_props = graph.query("""
                                        PREFIX owl: <http://www.w3.org/2002/07/owl#>
                                        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                                        SELECT DISTINCT ?c ?p WHERE {
                                            {?c rdfs:subClassOf [ owl:onProperty ?p ] }
                                            UNION
                                            {?p rdfs:domain ?c}
                                            FILTER (isURI(?p) && isURI(?c))
                                          }""")

    return set([__q_name(graph.namespace_manager, r.p) for r in all_class_props if
                __q_name(graph.namespace_manager, r.c) == ty])


def _get_type_references(graph, ty):
    # type: (Graph, str) -> set
    all_class_props = graph.query("""
                                    PREFIX owl: <http://www.w3.org/2002/07/owl#>
                                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                                    SELECT ?c ?p WHERE {
                                        { ?r owl:onProperty ?p .
                                          {?r owl:someValuesFrom ?c}
                                          UNION
                                          {?r owl:allValuesFrom ?c}
                                          UNION
                                          {?r owl:onClass ?c}
                                        }
                                        UNION
                                        {?p rdfs:range ?c}
                                        FILTER (isURI(?p) && isURI(?c))
                                       }""")

    all_types = __extend_with(_get_supertypes, graph, ty)
    return set([__q_name(graph.namespace_manager, r.p) for r in all_class_props if
                __q_name(graph.namespace_manager, r.c) in all_types])


def _get_type_specific_references(graph, ty):
    # type: (Graph, str) -> set
    all_class_props = graph.query("""
                                    PREFIX owl: <http://www.w3.org/2002/07/owl#>
                                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                                    SELECT ?c ?p WHERE {
                                        { ?r owl:onProperty ?p .
                                          {?r owl:someValuesFrom ?c}
                                          UNION
                                          {?r owl:allValuesFrom ?c}
                                          UNION
                                          {?r owl:onClass ?c}
                                        }
                                        UNION
                                        {?p rdfs:range ?c}
                                        FILTER (isURI(?p) && isURI(?c))
                                       }""")

    return set([__q_name(graph.namespace_manager, r.p) for r in all_class_props if
                __q_name(graph.namespace_manager, r.c) == ty])


def _context(f):
    # type: (callable) -> callable
    def wrap(self=None, *args, **kwargs):
        if self.cache is not None:
            return cached(self.cache)(f)(self, *args, **kwargs)
        else:
            return f(self, *args, **kwargs)

    return wrap


@_context
def _query(graph, q):
    return graph.query(q)


class SchemaGraph(object):
    def __init__(self, query_f, g):
        self.g = g
        self.__query = query_f

    def __getattr__(self, item):
        try:
            return self.__getattribute__(item)
        except:
            return self.g.__getattribute__(item)

    def query(self, q):
        return self.__query(self.g, q)

    def get_context(self, c):
        return SchemaGraph(self.__query, self.g.get_context(c))


class Schema(object):
    def __init__(self, cache=True):
        self.__cache = Cache() if cache else None
        self.__graph = None
        self.__namespaces = {}
        self.__prefixes = {}

    @property
    def cache(self):
        # type: () -> Cache
        return self.__cache

    @property
    def graph(self):
        # type: () -> ContextGraph
        return SchemaGraph(self.__query, self.__graph)

    @graph.setter
    def graph(self, g):
        self.__graph = g
        self.__graph.store.graph_aware = False
        self.update_ns_dicts()

    def update_ns_dicts(self):
        self.__namespaces.update([(uri, prefix) for (prefix, uri) in self.__graph.namespaces()])
        self.__prefixes.update([(prefix, uri) for (prefix, uri) in self.__graph.namespaces()])
        self.__clear_cache()

    def add_context(self, id, context):
        # type: (str, Graph) -> None
        _add_context(self.graph, id, context)
        self.update_ns_dicts()

    def update_context(self, id, context):
        # type: (str, Graph) -> None
        _update_context(self.graph, id, context)
        self.update_ns_dicts()

    def remove_context(self, id):
        # type: (str) -> None
        _remove_context(self.graph, id)
        self.update_ns_dicts()
        self.__clear_cache()

    def __clear_cache(self):
        if self.__cache is not None:
            self.__cache.clear()

    @property
    def contexts(self):
        # type: () -> iter
        return _contexts(self.graph)

    def get_context(self, id):
        # type: (str) -> Graph
        return _get_context(self.graph, id)

    @property
    def prefixes(self):
        # type: () -> dict
        return dict([(prefix, uri) for (prefix, uri) in self.__graph.namespaces()])

    @_context
    def get_types(self, context=None):
        # type: (object) -> iter
        if not isinstance(context, ContextGraph):
            context = self.graph.get_context(context) if context is not None else self.graph
        return _get_types(context)

    @_context
    def get_properties(self, context=None):
        # type: (object) -> iter
        if not isinstance(context, ContextGraph):
            context = self.graph.get_context(context) if context is not None else self.graph
        return _get_properties(context)

    @_context
    def is_object_property(self, p):
        # type: (str, Graph) -> bool
        return _is_object_property(self.graph, p)

    @_context
    def get_property_domain(self, p):
        # type: (str, Graph) -> iter
        return _get_property_domain(self.graph, p)

    @_context
    def get_property_range(self, p):
        # type: (str, Graph) -> iter
        return _get_property_range(self.graph, p)

    @_context
    def get_property_inverses(self, p):
        # type: (str, Graph) -> iter
        return _get_property_inverses(self.graph, p)

    @_context
    def get_property_constraints(self, p):
        # type: (str, Graph) -> iter
        return _get_property_constraints(self.graph, p)

    @_context
    def get_supertypes(self, t):
        # type: (str, Graph) -> iter
        return _get_supertypes(self.graph, t)

    @_context
    def get_subtypes(self, t):
        # type: (str, Graph) -> iter
        return _get_subtypes(self.graph, t)

    @_context
    def get_type_properties(self, t):
        # type: (str, Graph) -> iter
        return _get_type_properties(self.graph, t)

    @_context
    def get_type_specific_properties(self, t):
        # type: (str, Graph) -> iter
        return _get_type_specific_properties(self.graph, t)

    @_context
    def get_type_references(self, t):
        # type: (str, Graph) -> iter
        return _get_type_references(self.graph, t)

    @_context
    def get_type_specific_references(self, t):
        # type: (str, Graph) -> iter
        return _get_type_specific_references(self.graph, t)

    @_context
    def __query(self, g, q):
        return g.query(q)

    def close(self):
        pass
